""" dictyExpress widget """
from typing import Optional

from requests.exceptions import ConnectionError

from AnyQt.QtGui import QFont
from AnyQt.QtCore import Qt, QSize
from AnyQt.QtWidgets import QLabel, QTreeWidget, QTreeWidgetItem

from Orange.data import Table, StringVariable
from Orange.widgets import gui, settings
from Orange.widgets.widget import Msg, OWWidget
from Orange.widgets.utils.signals import Output
from Orange.widgets.utils.concurrent import ConcurrentWidgetMixin

from orangecontrib.bioinformatics import resolwe
from orangecontrib.bioinformatics.resolwe import genapi, connect
from orangecontrib.bioinformatics.ncbi.gene import ENTREZ_ID, GeneMatcher
from orangecontrib.bioinformatics.resolwe.utils import etc_to_table
from orangecontrib.bioinformatics.widgets.utils.data import (
    TAX_ID,
    GENE_ID_COLUMN,
    GENE_ID_ATTRIBUTE,
    GENE_AS_ATTRIBUTE_NAME,
)
from orangecontrib.bioinformatics.widgets.components.resolwe import SignIn, get_credential_manager

Labels = [
    (" ", " "),
    ("var.project", "Project"),
    ("static.name", "Experiment"),
    ("static.cite", "Citation"),
    ("var.growth", "Growth"),
    ("var.treatment", "Treatment"),
    ("var.strain", "Strain"),
]


class OWdictyExpress(OWWidget, ConcurrentWidgetMixin):

    name = "dictyExpress"
    description = "Time-course gene expression data"
    icon = "../widgets/icons/OWdictyExpress.svg"
    want_main_area = True
    priority = 20

    class Inputs:
        pass

    class Outputs:
        etc_data = Output("Data", Table)

    class Error(OWWidget.Error):
        unreachable_host = Msg('Host not reachable')
        invalid_credentials = Msg('Invalid credentials')

    gene_as_attr_name = settings.Setting(0)

    selected_item = settings.Setting(None, schema_only=True)
    auto_commit = settings.Setting(False, schema_only=True)

    def __init__(self):
        super().__init__()
        ConcurrentWidgetMixin.__init__(self)

        self._res: Optional[genapi.GenAPI] = None
        self.organism = '44689'
        self.server = 'https://dictyexpress.research.bcm.edu'
        self.headerLabels = [x[1] for x in Labels]
        self.searchString = ""
        self.items = []
        self.genapi_pub_auth = {
            'url': genapi.DEFAULT_URL,
            'username': genapi.DEFAULT_EMAIL,
            'password': genapi.DEFAULT_PASSWD,
        }

        # Login Section
        box = gui.widgetBox(self.controlArea, 'Sign in')
        self.user_info = gui.label(box, self, '')
        self.server_info = gui.label(box, self, '')

        box = gui.widgetBox(box, orientation=Qt.Horizontal)
        self.sign_in_btn = gui.button(box, self, 'Sign In', callback=self.sign_in, autoDefault=False)
        self.sign_out_btn = gui.button(box, self, 'Sign Out', callback=self.sign_out, autoDefault=False)

        box = gui.widgetBox(self.controlArea, "Output")
        gui.radioButtonsInBox(
            box, self, "gene_as_attr_name", ["Genes in rows", "Genes in columns"], callback=self.invalidate
        )

        self.clear_cache_btn = gui.button(
            self.controlArea, self, "Clear cache", autoDefault=False, callback=self.clear_cache
        )

        gui.rubber(self.controlArea)

        self.commit_button = gui.auto_commit(self.controlArea, self, "auto_commit", "&Commit", box=False)

        # Experiment Section

        label = QLabel("Available projects:")
        my_font = QFont()
        my_font.setBold(True)
        label.setFont(my_font)
        self.mainArea.layout().addWidget(label)

        self.filter = gui.lineEdit(
            self.mainArea, self, "searchString", "Filter:", callbackOnType=True, callback=self.search_update
        )

        self.experimentsWidget = QTreeWidget(
            alternatingRowColors=True, rootIsDecorated=False, uniformRowHeights=True, sortingEnabled=True
        )

        self.experimentsWidget.setItemDelegateForColumn(0, gui.IndicatorItemDelegate(self, role=Qt.DisplayRole))

        self.experimentsWidget.selectionModel().selectionChanged.connect(self.on_selection_changed)

        self.experimentsWidget.setHeaderLabels(self.headerLabels)
        self.mainArea.layout().addWidget(self.experimentsWidget)

        self.sign_in(silent=True)
        self.sizeHint()

    def sizeHint(self):
        return QSize(1400, 680)

    @property
    def res(self):
        return self._res

    @res.setter
    def res(self, value: genapi.GenAPI):
        if isinstance(value, genapi.GenAPI):
            self._res = value
            self.Error.clear()
            self.reset()
            self.load_experiments()
            self.update_user_status()
            self.Outputs.etc_data.send(None)

    def sign_in(self, silent=False):
        dialog = SignIn(self, server_type='genesis')

        if silent:
            dialog.sign_in()
            if dialog.resolwe_instance is not None:
                self.res = dialog.resolwe_instance
            else:
                self.res = connect(**self.genapi_pub_auth, server_type=resolwe.GENESIS_PLATFORM)

        if not silent and dialog.exec_():
            self.res = dialog.resolwe_instance

    def sign_out(self):
        # Remove username and password
        cm = get_credential_manager(resolwe.GENESIS_PLATFORM)
        del cm.username
        del cm.password
        # Use public credentials when user signs out
        self.res = connect(**self.genapi_pub_auth, server_type=resolwe.GENESIS_PLATFORM)

    def update_user_status(self):
        cm = get_credential_manager(resolwe.GENESIS_PLATFORM)

        if cm.username:
            user_info = f"User: {cm.username}"
            self.sign_in_btn.setEnabled(False)
            self.sign_out_btn.setEnabled(True)
        else:
            user_info = 'User: Anonymous'
            self.sign_in_btn.setEnabled(True)
            self.sign_out_btn.setEnabled(False)

        self.user_info.setText(user_info)
        self.server_info.setText(f'Server: {self.res._gen.url[8:]}')

    def clear_cache(self):
        resolwe.GenAPI.clear_cache()
        self.reset()
        self.load_experiments()

    def reset(self):
        self.experimentsWidget.clear()  # clear QTreeWidget
        self.items = []
        # self.lastSelected = None
        self.searchString = ""

    def search_update(self):
        parts = self.searchString.split()
        for item in self.items:
            item.setHidden(not all(s in item for s in parts))

    def on_exception(self, ex):
        if isinstance(ex, ConnectionError) or isinstance(ex, ValueError):
            self.Error.unreachable_host()

        print(ex)

    def on_done(self, results):
        if isinstance(results, list):
            self.load_tree_items(results)
        elif isinstance(results, tuple):
            self.send_to_output(results)

    def load_experiments(self):
        if self.res:
            self.start(self.res.fetch_etc_objects)

    def load_tree_items(self, list_of_exp):
        self.items = [CustomTreeItem(self.experimentsWidget, item) for item in list_of_exp]

        for i in range(len(self.headerLabels)):
            self.experimentsWidget.resizeColumnToContents(i)

        self.set_cached_indicator()
        self.set_selected()

    def set_selected(self):
        for item in self.items:
            if self.selected_item and item.gen_data_id == self.selected_item:
                self.experimentsWidget.setCurrentItem(item)

    def on_selection_changed(self):
        self.invalidate()

    def invalidate(self):
        self.commit()

    def send_to_output(self, result):
        etc_json, table_name = result

        # convert to table
        data = etc_to_table(etc_json, bool(self.gene_as_attr_name))
        # set table name
        data.name = table_name

        # match genes
        gene_matcher = GeneMatcher(str(self.organism))

        if not bool(self.gene_as_attr_name):
            if 'Gene' in data.domain:
                data = gene_matcher.match_table_column(data, 'Gene', StringVariable(ENTREZ_ID))
            data.attributes[GENE_ID_COLUMN] = ENTREZ_ID
        else:
            data = gene_matcher.match_table_attributes(data)
            data.attributes[GENE_ID_ATTRIBUTE] = ENTREZ_ID

        # add table attributes
        data.attributes[TAX_ID] = str(self.organism)
        data.attributes[GENE_AS_ATTRIBUTE_NAME] = bool(self.gene_as_attr_name)

        # reset cache indicators
        self.set_cached_indicator()
        # send data to the output signal
        self.Outputs.etc_data.send(data)

    def commit(self):
        self.Error.clear()
        selected_items = self.experimentsWidget.selectedItems()  # get selected TreeItem

        if len(selected_items) < 1:
            self.Outputs.etc_data.send(None)
            return

        selected_item = selected_items[0]
        self.selected_item = selected_item.gen_data_id
        self.start(self.res.download_etc_data, selected_item.gen_data_id, table_name=selected_item.data_name)

    def set_cached_indicator(self):
        cached = self.res.get_cached_ids()

        for item in self.items:

            if item.gen_data_id in cached:
                item.setData(0, Qt.DisplayRole, " ")
            else:
                item.setData(0, Qt.DisplayRole, "")


class CustomTreeItem(QTreeWidgetItem):
    def __init__(self, parent, gen_data):
        super(CustomTreeItem, self).__init__(parent)  # Init super class (QtGui.QTreeWidgetItem )

        self._gen_data = gen_data  # GenData object
        self.set_rows(self._gen_data.annotation)  # set rows in QTreeWidget

    def __contains__(self, text):
        return any(text.upper() in str(self.text(i)).upper() for i in range(self.columnCount()))

    @property
    def gen_data_id(self):
        return self._gen_data.id

    @property
    def data_name(self):
        try:
            project = self._gen_data.var['project']
            experiment = self._gen_data.static['name']
        except (AttributeError, KeyError):
            project = ''
            experiment = ''

        return '{} ({})'.format(project, experiment)

    def set_rows(self, row):
        for index, label in enumerate(Labels):
            if index > 0:
                try:
                    if type(row[label[0]]["value"]) == list:
                        self.setText(index, row[label[0]]["value"][0]["name"])
                    else:
                        self.setText(index, row[label[0]]["value"])
                except (IndexError, KeyError):
                    self.setText(index, 'No data')


if __name__ == "__main__":
    from orangewidget.utils.widgetpreview import WidgetPreview

    WidgetPreview(OWdictyExpress).run()



DOMAIN = 'gene'

gene_info_attributes = (
    'species',
    'tax_id',
    'gene_id',
    'symbol',
    'synonyms',
    'db_refs',
    'description',
    'locus_tag',
    'chromosome',
    'map_location',
    'type_of_gene',
    'symbol_from_nomenclature_authority',
    'full_name_from_nomenclature_authority',
    'nomenclature_status',
    'other_designations',
    'modification_date',
    'homology_group_id',
    'homologs',
)

_select_gene_info_columns = """
    gene_info.species, gene_info.tax_id, gene_info.gene_id, gene_info.symbol, gene_info.synonyms,  gene_info.db_refs,
    gene_info.description, gene_info.locus_tag, gene_info.chromosome,  gene_info.map_location, gene_info.type_of_gene,
    gene_info.symbol_from_nomenclature_authority, gene_info.full_name_from_nomenclature_authority,
    gene_info.nomenclature_status, gene_info.other_designations, gene_info.modification_date,
    gene_info.homology_group_id, gene_info.homologs
    """

query_exact = f"""
    SELECT  {_select_gene_info_columns}
    FROM gene_info
    JOIN gene_info_fts on gene_info.rowid = gene_info_fts.rowid
    WHERE gene_info_fts
        MATCH ?
            AND (lower(gene_info_fts.gene_id)=?
             OR  lower(gene_info_fts.symbol)=?
             OR  lower(gene_info_fts.locus_tag)=?
             OR  lower(gene_info_fts.symbol_from_nomenclature_authority)=?)
"""

query = f"""
    SELECT {_select_gene_info_columns}
    FROM gene_info
    JOIN gene_info_fts on gene_info.rowid = gene_info_fts.rowid
    WHERE gene_info_fts MATCH ?
"""

# Pretty strings
ENTREZ_ID = 'Entrez ID'
ENSEMBl_ID = 'Ensembl ID'
GENE_SYMBOL = 'Symbol'
GENE_SYNONYMS = 'Synonyms'



""" Gene Expression Omnibus datasets widget """
import sys
from types import SimpleNamespace
from typing import Any, Optional, DefaultDict
from collections import defaultdict

import requests

from AnyQt.QtCore import Qt
from AnyQt.QtWidgets import (
    QSplitter,
    QTreeWidget,
    QTreeWidgetItem,
    QAbstractItemView,
    QAbstractScrollArea,
)

from orangewidget.utils.itemdelegates import DataDelegate

from Orange.data import Table
from Orange.widgets.gui import (
    LinkRole,
    IndicatorItemDelegate,
    rubber,
    lineEdit,
    separator,
    widgetBox,
    auto_commit,
    widgetLabel,
    radioButtonsInBox,
)
from Orange.widgets.widget import Msg, OWWidget
from Orange.widgets.settings import Setting
from Orange.widgets.utils.signals import Output
from Orange.widgets.utils.tableview import TableView
from Orange.widgets.utils.concurrent import TaskState, ConcurrentWidgetMixin

from orangecontrib.bioinformatics.geo import is_cached, pubmed_url, local_files
from orangecontrib.bioinformatics.geo.dataset import (
    GDSInfo,
    get_samples,
    dataset_download,
)
from orangecontrib.bioinformatics.widgets.utils.gui import FilterProxyModel
from orangecontrib.bioinformatics.widgets.utils.itemmodels import TableModel
from orangecontrib.bioinformatics.widgets.utils.itemdelegates import (
    LinkStyledItemDelegate,
)


class Result(SimpleNamespace):
    gds_dataset: Optional[Table] = None


def run_download_task(
    gds_id: str, samples: DefaultDict[str, list], transpose: bool, state: TaskState
):
    res = Result()
    current_iter = 0
    max_iter = 102

    def callback():
        nonlocal current_iter
        current_iter += 1
        state.set_progress_value(100 * (current_iter / max_iter))

    state.set_status("Downloading...")
    res.gds_dataset = dataset_download(
        gds_id, samples, transpose=transpose, callback=callback
    )
    return res


class GEODatasetsModel(TableModel):
    (
        indicator_col,
        gds_id_col,
        pubmed_id_col,
        organism_col,
        samples_col,
        features_col,
        genes_col,
        subsets_col,
        title_col,
    ) = range(9)

    def __init__(self, gds: GDSInfo):
        items = list(gds.values())
        pubmed_ids = [g.get("pubmed_id", None) for g in items]

        def title(name: str):
            return {Qt.DisplayRole: name}

        def pubmed_id(row: int) -> Optional[str]:
            pubmed = pubmed_ids[row]
            if isinstance(pubmed, list):
                return pubmed[0] if len(pubmed) > 0 else None
            else:
                return pubmed

        def render_pubmed_url(row):
            id_ = pubmed_id(row)
            if id_:
                return pubmed_url.format(id_)
            return None

        columns = [
            TableModel.Column(
                title(""),
                {
                    Qt.DisplayRole: lambda row: " "
                    if is_cached(items[row]["name"])
                    else "",
                    Qt.UserRole: lambda row: items[row],
                },
            ),
            TableModel.Column(
                title("ID"),
                {
                    Qt.DisplayRole: lambda row: items[row]["name"],
                    Qt.UserRole: lambda row: items[row],
                },
            ),
            TableModel.Column(
                title("PubMedID"),
                {
                    Qt.DisplayRole: pubmed_id,
                    LinkRole: render_pubmed_url,
                    Qt.ToolTipRole: render_pubmed_url,
                },
            ),
            TableModel.Column(
                title("Organism"),
                {
                    Qt.DisplayRole: lambda row: items[row]["sample_organism"],
                },
            ),
            TableModel.Column(
                title("Samples"),
                {
                    Qt.DisplayRole: lambda row: len(get_samples(items[row])),
                },
            ),
            TableModel.Column(
                title("Features"),
                {
                    Qt.DisplayRole: lambda row: items[row]["variables"],
                },
            ),
            TableModel.Column(
                title("Genes"),
                {
                    Qt.DisplayRole: lambda row: items[row]["genes"],
                },
            ),
            TableModel.Column(
                title("Subsets"),
                {
                    Qt.DisplayRole: lambda row: len(items[row]["subsets"]),
                },
            ),
            TableModel.Column(
                title("Title"),
                {
                    Qt.DisplayRole: lambda row: items[row]["title"],
                    Qt.ToolTipRole: lambda row: items[row]["title"],
                },
            ),
        ]
        super().__init__(len(items), columns)
        self.info = gds

    def sortColumnData(self, column):
        if column == self.gds_id_col:
            items = [g["name"] for g in self.info.values()]
            items = [int(name.strip('GDS')) for name in items]
            return items
        if column == self.pubmed_id_col:
            items = [
                self.data(self.index(i, column), Qt.DisplayRole)
                for i in self.mapFromSourceRows(range(self.rowCount()))
            ]
            items = [int(pid or 0) for pid in items]
            return items
        return super().sortColumnData(column)

    def update_cache_indicator(self):
        self.dataChanged.emit(self.index(0, 0), self.index(self.rowCount() - 1, 0))


class FilterProxyModel(FilterProxyModel):
    def sort(self, column: int, order: Qt.SortOrder = Qt.AscendingOrder) -> None:
        """
        Reimplemented.

        Dispatch the sorting to the source model.
        """
        source = self.sourceModel()
        if source is not None:
            source.sort(column, order)


class OWGEODatasets(OWWidget, ConcurrentWidgetMixin):
    name = "GEO Data Sets"
    description = "Access to Gene Expression Omnibus data sets."
    icon = "icons/OWGEODatasets.svg"
    priority = 10

    class Warning(OWWidget.Warning):
        using_local_files = Msg("Can't connect to serverfiles. Using cached files.")

    class Error(OWWidget.Error):
        no_connection = Msg("Widget can't connect to serverfiles.")

    class Outputs:
        gds_data = Output("Expression Data", Table)

    search_pattern = Setting('')
    auto_commit = Setting(True)
    genes_as_rows = Setting(False)
    selected_gds = Setting(None)
    gds_selection_states = Setting({})
    splitter_settings = Setting(
        (
            b'\x00\x00\x00\xff\x00\x00\x00\x00\x00\x00\x00\x02\x00\x00\x01'
            b'\xea\x00\x00\x00\xd7\x01\x00\x00\x00\x07\x01\x00\x00\x00\x02',
            b'\x00\x00\x00\xff\x00\x00\x00\x00\x00\x00\x00\x02\x00\x00\x01'
            b'\xb5\x00\x00\x02\x10\x01\x00\x00\x00\x07\x01\x00\x00\x00\x01',
        )
    )

    def __init__(self):
        OWWidget.__init__(self)
        ConcurrentWidgetMixin.__init__(self)

        try:
            self.gds_info: Optional[GDSInfo] = GDSInfo()
        except requests.exceptions.ConnectionError:
            self.gds_info = {}
            self.Error.no_connection()
            return

        self.gds_data: Optional[Table] = None
        self.__updating_filter = False

        # Control area
        box = widgetBox(self.controlArea, 'Info', addSpace=True)
        self.infoBox = widgetLabel(box, 'Initializing\n\n')

        box = widgetBox(self.controlArea, 'Output', addSpace=True)
        radioButtonsInBox(
            box,
            self,
            'genes_as_rows',
            ['Samples in rows', 'Genes in rows'],
            callback=self._run,
        )
        separator(box)

        rubber(self.controlArea)
        auto_commit(self.controlArea, self, 'auto_commit', '&Commit', box=False)

        # Main Area

        # Filter widget
        self.filter = lineEdit(
            self.mainArea,
            self,
            'search_pattern',
            'Filter:',
            callbackOnType=True,
            callback=self._apply_filter,
        )
        self.mainArea.layout().addWidget(self.filter)

        splitter_vertical = QSplitter(Qt.Vertical, self.mainArea)

        self.mainArea.layout().addWidget(splitter_vertical)

        # set table view
        self.table_view = TableView(
            splitter_vertical,
            showGrid=False,
            sortingEnabled=True,
            alternatingRowColors=True,
            selectionBehavior=QAbstractItemView.SelectRows,
            selectionMode=QAbstractItemView.SingleSelection,
            sizeAdjustPolicy=QAbstractScrollArea.AdjustToContents,
        )
        self.table_view.sortByColumn(1, Qt.AscendingOrder)
        self.table_view.verticalHeader().setVisible(False)
        self.table_view.viewport().setMouseTracking(True)

        self.table_model = GEODatasetsModel(self.gds_info)
        self.proxy_model = FilterProxyModel()
        self.proxy_model.setSourceModel(self.table_model)
        self.table_view.setModel(self.proxy_model)

        self.table_view.horizontalHeader().setStretchLastSection(True)
        self.table_view.resizeColumnsToContents()

        # set item delegates
        self.table_view.setItemDelegate(DataDelegate(self.table_view))
        self.table_view.setItemDelegateForColumn(
            self.table_model.pubmed_id_col, LinkStyledItemDelegate(self.table_view)
        )
        self.table_view.setItemDelegateForColumn(
            self.table_model.gds_id_col, LinkStyledItemDelegate(self.table_view)
        )
        self.table_view.setItemDelegateForColumn(
            self.table_model.indicator_col,
            IndicatorItemDelegate(self.table_view, role=Qt.DisplayRole),
        )

        splitter_horizontal = QSplitter(Qt.Horizontal, splitter_vertical)

        # Description Widget
        box = widgetBox(splitter_horizontal, 'Description')
        self.description_widget = widgetLabel(box, '')
        self.description_widget.setWordWrap(True)
        rubber(box)

        # Sample Annotations Widget
        box = widgetBox(splitter_horizontal, 'Sample Annotations')
        self.annotations_widget = QTreeWidget(box)
        self.annotations_widget.setHeaderLabels(
            ['Type (Sample annotations)', 'Sample count']
        )
        self.annotations_widget.setRootIsDecorated(True)
        box.layout().addWidget(self.annotations_widget)
        self._annotations_updating = False
        self.annotations_widget.itemChanged.connect(
            self.on_annotation_selection_changed
        )
        self.splitters = splitter_vertical, splitter_horizontal

        for sp, setting in zip(self.splitters, self.splitter_settings):
            sp.splitterMoved.connect(self._splitter_moved)
            sp.restoreState(setting)

        self.table_view.selectionModel().selectionChanged.connect(
            self.__on_selection_changed
        )
        self._apply_filter()

        self.commit()

    def _splitter_moved(self, *args):
        self.splitter_settings = [bytes(sp.saveState()) for sp in self.splitters]

    def _set_description_widget(self):
        self.description_widget.setText(
            self.selected_gds.get('description', 'Description not available.')
        )

    def _set_annotations_widget(self, gds):
        self._annotations_updating = True
        self.annotations_widget.clear()

        annotations = defaultdict(set)
        subsets_count = {}

        for desc in gds['subsets']:
            annotations[desc['type']].add(desc['description'])
            subsets_count[desc['description']] = str(len(desc['sample_id']))

        for _type, subsets in annotations.items():
            key = (gds["name"], _type)
            parent = QTreeWidgetItem(self.annotations_widget, [_type])
            parent.key = key
            for subset in subsets:
                key = (gds['name'], _type, subset)
                item = QTreeWidgetItem(parent, [subset, subsets_count.get(subset, '')])
                item.setFlags(item.flags() | Qt.ItemIsUserCheckable)
                item.setCheckState(0, self.gds_selection_states.get(key, Qt.Checked))
                item.key = key

        self._annotations_updating = False
        self.annotations_widget.expandAll()
        for i in range(self.annotations_widget.columnCount()):
            self.annotations_widget.resizeColumnToContents(i)

    def _set_selection(self):
        if self.selected_gds is not None:
            index = None
            ids = self.proxy_model.match(
                self.proxy_model.index(0, 1),
                Qt.DisplayRole,
                self.selected_gds["name"],
                1,
                Qt.MatchCaseSensitive | Qt.MatchExactly,
            )
            if ids:
                index = ids[0]

            if index is not None:
                self.table_view.selectionModel().blockSignals(True)
                self.table_view.selectRow(index.row())
                self._handle_selection_changed()
                self.table_view.selectionModel().blockSignals(False)

    def _handle_selection_changed(self):
        if self.table_model is not None:
            selection = self.table_view.selectionModel().selectedRows(
                self.table_model.gds_id_col
            )
            selected_gds_name = selection[0].data() if len(selection) > 0 else None

            if selected_gds_name:
                self.selected_gds = self.table_model.info.get(selected_gds_name)
                self._set_annotations_widget(self.selected_gds)
                self._set_description_widget()
            else:
                self.annotations_widget.clear()
                self.description_widget.clear()

            self.update_info()

    def _apply_filter(self):
        keys = self.search_pattern.lower().split()

        def filter(gds):
            source = (gds["name"], gds["title"], gds["sample_organism"])
            source = [t.lower() for t in source]
            return all(any(k in s for s in source) for k in keys)

        changed = False

        def current_changed():
            nonlocal changed
            changed = True

        selmodel = self.table_view.selectionModel()
        selmodel.selectionChanged.connect(current_changed)
        try:
            self.__updating_filter = True
            self.proxy_model.set_filters(
                [
                    FilterProxyModel.Filter(
                        self.table_model.gds_id_col, Qt.UserRole, filter
                    )
                ]
            )
            if changed:
                self.table_view.clearSelection()
                self._handle_selection_changed()
        finally:
            selmodel.selectionChanged.disconnect(current_changed)
            self.__updating_filter = False

        self._set_selection()

    def _run(self):
        self.Warning.using_local_files.clear()
        if self.selected_gds is not None:
            self.gds_data = None
            self.start(
                run_download_task,
                self.selected_gds.get('name'),
                self.get_selected_samples(),
                self.genes_as_rows,
            )

    def __on_selection_changed(self):
        if not self.__updating_filter:
            self.on_gds_selection_changed()

    def on_gds_selection_changed(self):
        self._handle_selection_changed()
        self.commit()

    def on_annotation_selection_changed(self):
        if self._annotations_updating:
            return
        for i in range(self.annotations_widget.topLevelItemCount()):
            item = self.annotations_widget.topLevelItem(i)
            if 'key' in item.__dict__:
                self.gds_selection_states[item.key] = item.checkState(0)
            for j in range(item.childCount()):
                child = item.child(j)
                if 'key' in child.__dict__:
                    self.gds_selection_states[child.key] = child.checkState(0)

        self.commit()

    def update_info(self):
        all_gds = len(self.table_model.info)
        text = "{} datasets\n{} datasets cached\n".format(
            all_gds, len(local_files.listfiles())
        )
        filtered = self.table_view.model().rowCount()
        if all_gds != filtered:
            text += "{} after filtering".format(filtered)
        self.infoBox.setText(text)

    def get_selected_samples(self):
        """
        Return the currently selected sample annotations.

        The return value is a list of selected (sample type, sample value)
        tuples.

        .. note:: if some Sample annotation type has no selected values.
                  this method will return all values for it.

        TODO: this could probably be simplified.

        """

        def childiter(item):
            """Iterate over the children of an QTreeWidgetItem instance."""
            for i in range(item.childCount()):
                yield item.child(i)

        samples = []
        unused_types = []
        used_types = []

        for stype in childiter(self.annotations_widget.invisibleRootItem()):
            selected_values = []
            all_values = []
            for sval in childiter(stype):
                value = (str(stype.text(0)), str(sval.text(0)))
                if self.gds_selection_states.get(sval.key, True):
                    selected_values.append(value)
                all_values.append(value)
            if selected_values:
                samples.extend(selected_values)
                used_types.append(str(stype.text(0)))
            else:
                # If no sample of sample type is selected we don't filter on it.
                samples.extend(all_values)
                unused_types.append(str(stype.text(0)))

        _samples = defaultdict(list)
        for sample, sample_type in samples:
            _samples[sample].append(sample_type)
        return _samples

    def commit(self):
        self._run()

    def on_exception(self, ex: Exception):
        self.Warning.using_local_files()

    def on_done(self, result: Result):
        assert isinstance(result.gds_dataset, Table)
        self.gds_data = result.gds_dataset

        if self.gds_info:
            self.table_model.update_cache_indicator()

        self.Outputs.gds_data.send(self.gds_data)

    def on_partial_result(self, result: Any) -> None:
        pass

    def onDeleteWidget(self):
        self.shutdown()
        super().onDeleteWidget()

    def send_report(self):
        self.report_items(
            "GEO Dataset",
            [
                ("ID", self.selected_gds['name']),
                ("Title", self.selected_gds['title']),
                ("Organism", self.selected_gds['sample_organism']),
            ],
        )
        self.report_items(
            "Data",
            [
                ("Samples", self.selected_gds['sample_count']),
                ("Features", self.selected_gds['variables']),
                ("Genes", self.selected_gds['genes']),
            ],
        )
        self.report_name("Sample annotations")
        subsets = defaultdict(list)
        for subset in self.selected_gds['subsets']:
            subsets[subset['type']].append(
                (subset['description'], len(subset['sample_id']))
            )
        self.report_html += "<ul>"
        for _type in subsets:
            self.report_html += "<b>" + _type + ":</b></br>"
            for desc, count in subsets[_type]:
                self.report_html += 9 * "&nbsp" + "<b>{}:</b> {}</br>".format(
                    desc, count
                )
        self.report_html += "</ul>"


if __name__ == "__main__":

    def main_test():
        from AnyQt.QtWidgets import QApplication

        app = QApplication([])
        w = OWGEODatasets()
        w.show()
        w.raise_()
        r = app.exec_()
        w.saveSettings()
        return r

    sys.exit(main_test())



""" Qt component for Gene sets """
from typing import Union
from collections import defaultdict

import numpy as np

from AnyQt.QtCore import Qt
from AnyQt.QtWidgets import QWidget, QGroupBox, QTreeView, QTreeWidget, QTreeWidgetItem, QTreeWidgetItemIterator

from orangecontrib.bioinformatics.geneset import GeneSet, GeneSets, list_all, load_gene_sets

# TODO: better handle stored selection
# TODO: Don't use hardcoded 'Custom sets', use table name if available


class GeneSetsSelection(QWidget):
    def __init__(self, box, parent, settings_var, **kwargs):
        # type: (Union[QGroupBox, QWidget], QWidget, str) -> None
        super().__init__(**kwargs)

        self.parent = parent
        self.stored_selection = settings_var
        # gene sets object
        self.gs_object = GeneSets()  # type: GeneSets

        self.hierarchy_tree_widget = QTreeWidget(self)
        self.hierarchy_tree_widget.setHeaderHidden(True)
        self.hierarchy_tree_widget.setEditTriggers(QTreeView.NoEditTriggers)
        box.layout().addWidget(self.hierarchy_tree_widget)

        self.custom_set_hier = None
        self.default_selection = [
            ('GO', 'molecular_function'),
            ('GO', 'biological_process'),
            ('GO', 'cellular_component'),
        ]

    def clear_custom_sets(self):
        # delete any custom sets if they exists
        self.gs_object.delete_sets_by_hierarchy(self.custom_set_hier)

    def add_custom_sets(self, gene_sets_names, gene_names, hierarchy_title=None, select_customs_flag=False):
        # type: (np.ndarray, np.ndarray) -> None

        self.custom_set_hier = hierarchy_title
        self.clear_custom_sets()

        temp_dict = defaultdict(list)
        for set_name, gene_name in zip(gene_sets_names, gene_names):
            temp_dict[set_name].append(gene_name)

        g_sets = []
        for key, value in temp_dict.items():
            g_sets.append(
                GeneSet(
                    gs_id=key,
                    hierarchy=self.custom_set_hier,
                    organism=self.gs_object.common_org(),
                    name=key,
                    genes=set(value),
                )
            )

        self.gs_object.update(g_sets)
        self.update_gs_hierarchy(select_customs_flag=select_customs_flag)

    def load_gene_sets(self, tax_id):
        # type: (str) -> None
        self.gs_object = GeneSets()
        self.clear()

        gene_sets = list_all(organism=tax_id)
        self.set_hierarchy_model(self.hierarchy_tree_widget, self.hierarchy_tree(gene_sets))

        for gene_set in gene_sets:
            g_sets = load_gene_sets(gene_set, tax_id)
            self.gs_object.update(list(g_sets))

        self.set_selected_hierarchies()

    def clear_gene_sets(self):
        self.gs_object = GeneSets()

    def clear(self):
        # reset hierarchy widget state
        self.hierarchy_tree_widget.clear()

    def update_gs_hierarchy(self, select_customs_flag=False):
        self.clear()
        self.set_hierarchy_model(self.hierarchy_tree_widget, self.hierarchy_tree(self.gs_object.hierarchies()))
        if select_customs_flag:
            self.set_custom_sets()
        else:
            self.set_selected_hierarchies()

    def set_hierarchy_model(self, tree_widget, sets):
        def beautify_displayed_text(text):
            if '_' in text:
                return text.replace('_', ' ').title()
            else:
                return text

        # TODO: maybe optimize this code?
        for key, value in sets.items():
            item = QTreeWidgetItem(tree_widget, [beautify_displayed_text(key)])
            item.setFlags(item.flags() & (Qt.ItemIsUserCheckable | Qt.ItemIsSelectable | Qt.ItemIsEnabled))
            item.setExpanded(True)
            item.hierarchy = key

            if value:
                item.setFlags(item.flags() | Qt.ItemIsUserTristate)
                self.set_hierarchy_model(item, value)
            else:
                if item.parent():
                    item.hierarchy = (item.parent().hierarchy, key)

            if not item.childCount() and not item.parent():
                item.hierarchy = (key,)

    def get_hierarchies(self, **kwargs):
        """return selected hierarchy"""
        only_selected = kwargs.get('only_selected', None)

        sets_to_display = []

        if only_selected:
            iterator = QTreeWidgetItemIterator(self.hierarchy_tree_widget, QTreeWidgetItemIterator.Checked)
        else:
            iterator = QTreeWidgetItemIterator(self.hierarchy_tree_widget)

        while iterator.value():
            # note: if hierarchy value is not a tuple, then this is just top level qTreeWidgetItem that
            #       holds subcategories. We don't want to display all sets from category
            if type(iterator.value().hierarchy) is not str:

                if not only_selected:
                    sets_to_display.append(iterator.value().hierarchy)
                else:
                    if not iterator.value().isDisabled():
                        sets_to_display.append(iterator.value().hierarchy)

            iterator += 1

        return sets_to_display

    def set_selected_hierarchies(self):
        iterator = QTreeWidgetItemIterator(self.hierarchy_tree_widget, QTreeWidgetItemIterator.All)
        defaults = []

        while iterator.value():

            # note: if hierarchy value is not a tuple, then this is just top level qTreeWidgetItem that
            #       holds subcategories. We don't want to display all sets from category
            if type(iterator.value().hierarchy) is not str:
                if iterator.value().hierarchy in self.parent.__getattribute__(self.stored_selection):
                    iterator.value().setCheckState(0, Qt.Checked)
                else:
                    iterator.value().setCheckState(0, Qt.Unchecked)

            # if no items are checked, set defaults
            if iterator.value().hierarchy in self.default_selection:
                defaults.append(iterator.value())

            iterator += 1

        if len(self.get_hierarchies(only_selected=True)) == 0:
            [item.setCheckState(0, Qt.Checked) for item in defaults]

    def set_custom_sets(self):
        iterator = QTreeWidgetItemIterator(self.hierarchy_tree_widget, QTreeWidgetItemIterator.All)

        while iterator.value():

            # note: if hierarchy value is not a tuple, then this is just top level qTreeWidgetItem that
            #       holds subcategories. We don't want to display all sets from category
            if type(iterator.value().hierarchy) is not str:
                if iterator.value().hierarchy == self.custom_set_hier:
                    iterator.value().setCheckState(0, Qt.Checked)
                else:
                    iterator.value().setCheckState(0, Qt.Unchecked)

            iterator += 1

    @staticmethod
    def hierarchy_tree(gene_sets):
        def tree():
            return defaultdict(tree)

        collection = tree()

        def collect(col, set_hierarchy):
            if set_hierarchy:
                collect(col[set_hierarchy[0]], set_hierarchy[1:])

        for hierarchy in gene_sets:
            collect(collection, hierarchy)

        return collection



import unittest

import numpy as np

from AnyQt.QtCore import QModelIndex, QItemSelectionModel

from Orange.data import Table
from Orange.widgets.tests.base import WidgetTest
from Orange.widgets.tests.utils import simulate

from orangecontrib.bioinformatics.utils import serverfiles
from orangecontrib.bioinformatics.widgets.OWMarkerGenes import SERVER_FILES_DOMAIN, TreeItem, TreeView, OWMarkerGenes


class TestTreeItem(unittest.TestCase):
    def setUp(self) -> None:
        self.item1 = TreeItem("item 1", False, None, None)
        self.item2 = TreeItem("item 2", False, None, None)
        self.item3 = TreeItem("item 3", False, None, None)

    def test_change_parent(self):
        self.item1.change_parent(self.item2)
        self.assertEqual(self.item1.parentItem, self.item2)
        self.assertIn(self.item1, self.item2.childItems)

    def test_change_parent_none(self):
        self.item1.change_parent(None)
        self.assertEqual(self.item1.parentItem, None)
        self.assertListEqual([], self.item2.childItems)

    def change_parents(self):
        TreeItem.change_parents([self.item1, self.item2], self.item3)
        self.assertEqual(self.item1.parentItem, self.item3)
        self.assertIn(self.item1, self.item3.childItems)
        self.assertEqual(self.item2.parentItem, self.item3)
        self.assertIn(self.item2, self.item3.childItems)

    def test_append_child(self):
        self.item1.append_child(self.item3)
        self.assertIn(self.item3, self.item1.childItems)

    def test_remove_children(self):
        TreeItem.change_parents([self.item1, self.item2], self.item3)
        self.assertListEqual([self.item1, self.item2], self.item3.childItems)

        s = self.item3.remove_children(-1, 1)  # should return False (it is not possible)
        self.assertFalse(s)
        self.assertListEqual([self.item1, self.item2], self.item3.childItems)

        s = self.item3.remove_children(2, 1)  # should return False (it is not possible)
        self.assertFalse(s)
        self.assertListEqual([self.item1, self.item2], self.item3.childItems)

        s = self.item3.remove_children(1, 2)  # should return False (it is not possible)
        self.assertFalse(s)
        self.assertListEqual([self.item1, self.item2], self.item3.childItems)

        s = self.item3.remove_children(0, 1)
        self.assertTrue(s)
        self.assertListEqual([self.item2], self.item3.childItems)

        s = self.item3.remove_children(0, 1)
        self.assertTrue(s)
        self.assertListEqual([], self.item3.childItems)

        s = self.item3.remove_children(0, 1)
        self.assertFalse(s)
        self.assertListEqual([], self.item3.childItems)

        # test remove multiple
        TreeItem.change_parents([self.item1, self.item2], self.item3)
        s = self.item3.remove_children(0, 2)
        self.assertTrue(s)
        self.assertListEqual([], self.item3.childItems)

    def test_children_count(self):
        self.assertEqual(0, self.item3.children_count())

        TreeItem.change_parents([self.item1, self.item2], self.item3)
        self.assertEqual(2, self.item3.children_count())

        self.item3.remove_children(0, 1)
        self.assertEqual(1, self.item3.children_count())

        self.item3.remove_children(0, 1)
        self.assertEqual(0, self.item3.children_count())

    def test_row(self):
        # when no parent it is assumed that the item is a root item - its row is 0
        self.assertEqual(0, self.item1.row())
        self.assertEqual(0, self.item2.row())
        self.assertEqual(0, self.item3.row())

        # add parent
        TreeItem.change_parents([self.item1, self.item2], self.item3)
        self.assertEqual(0, self.item1.row())
        self.assertEqual(1, self.item2.row())
        self.assertEqual(0, self.item3.row())

        # different order
        self.item3.remove_children(0, 2)
        TreeItem.change_parents([self.item2, self.item1], self.item3)
        self.assertEqual(1, self.item1.row())
        self.assertEqual(0, self.item2.row())
        self.assertEqual(0, self.item3.row())

    def test_child_from_name_index(self):
        self.assertTupleEqual((None, None), self.item3.child_from_name_index("item 1"))
        TreeItem.change_parents([self.item1, self.item2], self.item3)

        self.assertTupleEqual((0, self.item1), self.item3.child_from_name_index("item 1"))
        self.assertTupleEqual((1, self.item2), self.item3.child_from_name_index("item 2"))

        self.item3.remove_children(0, 2)
        TreeItem.change_parents([self.item2, self.item1], self.item3)

        self.assertTupleEqual((1, self.item1), self.item3.child_from_name_index("item 1"))
        self.assertTupleEqual((0, self.item2), self.item3.child_from_name_index("item 2"))

    def test_contains_text(self):
        # simulate RowItem with dictionary
        self.item1.data_row = {"Name": "item1", "Cell Type": "type1", "Function": "function1"}
        self.item2.data_row = {"Name": "item2", "Cell Type": "type2", "Function": "function2"}
        filter_columns = ["Name", "Cell Type", "Function"]
        TreeItem.change_parents([self.item2, self.item1], self.item3)

        self.assertTrue(self.item1.contains_text("item1", filter_columns))
        self.assertFalse(self.item2.contains_text("item1", filter_columns))
        self.assertFalse(self.item3.contains_text("item1", filter_columns))

        self.assertTrue(self.item1.contains_text("type1", filter_columns))
        self.assertFalse(self.item2.contains_text("type1", filter_columns))
        self.assertFalse(self.item3.contains_text("type1", filter_columns))

        self.assertTrue(self.item1.contains_text("function1", filter_columns))
        self.assertFalse(self.item2.contains_text("function1", filter_columns))
        self.assertFalse(self.item3.contains_text("function1", filter_columns))

        self.assertFalse(self.item1.contains_text("item2", filter_columns))
        self.assertTrue(self.item2.contains_text("item2", filter_columns))
        self.assertFalse(self.item3.contains_text("item2", filter_columns))

        self.assertFalse(self.item1.contains_text("type2", filter_columns))
        self.assertTrue(self.item2.contains_text("type2", filter_columns))
        self.assertFalse(self.item3.contains_text("type2", filter_columns))

        self.assertFalse(self.item1.contains_text("function2", filter_columns))
        self.assertTrue(self.item2.contains_text("function2", filter_columns))
        self.assertFalse(self.item3.contains_text("function2", filter_columns))

        # test case sensitivity
        self.assertTrue(self.item1.contains_text("Item1", filter_columns))
        self.assertFalse(self.item2.contains_text("Item1", filter_columns))

        self.assertTrue(self.item1.contains_text("Type1", filter_columns))
        self.assertFalse(self.item2.contains_text("Type1", filter_columns))

        self.assertFalse(self.item1.contains_text("Item2", filter_columns))
        self.assertTrue(self.item2.contains_text("Item2", filter_columns))

        self.assertFalse(self.item1.contains_text("Type2", filter_columns))
        self.assertTrue(self.item2.contains_text("Type2", filter_columns))

    def test_get_data_rows(self):
        TreeItem.change_parents([self.item2, self.item1], self.item3)

        self.item1.data_row = "dr1"
        self.item2.data_row = "dr2"
        self.item3.data_row = "dr3"

        self.assertListEqual(sorted(["dr1", "dr2", "dr3"]), sorted(self.item3.get_data_rows()))
        self.assertListEqual(["dr1"], self.item1.get_data_rows())
        self.assertListEqual(["dr2"], self.item2.get_data_rows())

        self.item1.data_row = "dr1"
        self.item2.data_row = "dr2"
        self.item3.data_row = None

        self.assertListEqual(sorted(["dr1", "dr2"]), sorted(self.item3.get_data_rows()))
        self.assertListEqual(["dr1"], self.item1.get_data_rows())
        self.assertListEqual(["dr2"], self.item2.get_data_rows())

        self.item1.data_row = "dr1"
        self.item2.data_row = None
        self.item3.data_row = None

        self.assertListEqual(["dr1"], self.item3.get_data_rows())
        self.assertListEqual(["dr1"], self.item1.get_data_rows())
        self.assertListEqual([], self.item2.get_data_rows())

    def test_count_marker_genes(self):
        TreeItem.change_parents([self.item2, self.item1], self.item3)

        self.item1.data_row = "dr1"
        self.item2.data_row = "dr2"
        self.item3.data_row = "dr3"

        self.assertEqual(3, self.item3.count_marker_genes())
        self.assertEqual(1, self.item1.count_marker_genes())
        self.assertEqual(1, self.item2.count_marker_genes())

        self.item1.data_row = "dr1"
        self.item2.data_row = "dr2"
        self.item3.data_row = None

        self.assertEqual(2, self.item3.count_marker_genes())
        self.assertEqual(1, self.item1.count_marker_genes())
        self.assertEqual(1, self.item2.count_marker_genes())

        self.item1.data_row = "dr1"
        self.item2.data_row = None
        self.item3.data_row = None

        self.assertEqual(1, self.item3.count_marker_genes())
        self.assertEqual(1, self.item1.count_marker_genes())
        self.assertEqual(0, self.item2.count_marker_genes())


class TestOWMarkerGenes(WidgetTest):
    @classmethod
    def setUpClass(cls):
        """Code executed only once for all tests"""
        super().setUpClass()
        file_name = "panglao_gene_markers.tab"
        serverfiles.update(SERVER_FILES_DOMAIN, file_name)
        file_path = serverfiles.localpath_download(SERVER_FILES_DOMAIN, file_name)
        cls.panglao = Table.from_file(file_path)

        file_name = "cellMarker_gene_markers.tab"
        serverfiles.update(SERVER_FILES_DOMAIN, file_name)
        file_path = serverfiles.localpath_download(SERVER_FILES_DOMAIN, file_name)
        cls.cell_markers = Table.from_file(file_path)

    def test_minimum_size(self):
        pass

    def setUp(self):
        self.widget = self.create_widget(OWMarkerGenes)

    def move_top_element(self, source_view: TreeView) -> None:
        """
        This function moves top element form the source_view to another.

        Parameters
        ----------
        source_view
            The view from which we move top element.
        """
        parent_index = source_view.model().index(0, 0)
        source_view.selectionModel().select(parent_index, QItemSelectionModel.Select | QItemSelectionModel.Rows)
        self.widget.move_button.click()

    def move_first_child(self, source_view: TreeView) -> None:
        """
        This function moves the first child element from source_view to another.

        Parameters
        ----------
        source_view
            The view from which we move top element.
        """
        parent_index = source_view.model().index(0, 0)
        gene_index = source_view.model().index(0, 0, parent_index)
        source_view.selectionModel().select(gene_index, QItemSelectionModel.Select | QItemSelectionModel.Rows)
        self.widget.move_button.click()

    def test_data_not_empty(self):
        """
        Test that data are present
        """
        w = self.widget
        self.assertIsNotNone(w.data)

    def test_available_sources(self):
        """
        Test if all sources retrieved. Currently there are two sources available. When more available unittest will be
        changed.
        """
        self.assertListEqual(["CellMarker", "Panglao"], list(self.widget.available_sources.keys()))

    def test_source_changed(self):
        """
        Test changing the source of the data.
        """
        # cell marker data
        self.assertEqual("Panglao", self.widget.db_source_cb.currentText())
        self.assertTrue(isinstance(self.widget.data, Table))
        self.assertEqual(len(self.panglao), len(self.widget.data))

        # Panglao data
        simulate.combobox_activate_index(self.widget.controls.source_index, 1)
        self.assertEqual("CellMarker", self.widget.db_source_cb.currentText())
        self.assertTrue(isinstance(self.widget.data, Table))
        self.assertEqual(len(self.cell_markers), len(self.widget.data))

        # cell marker data
        simulate.combobox_activate_index(self.widget.controls.source_index, 0)
        self.assertEqual("Panglao", self.widget.db_source_cb.currentText())
        self.assertTrue(isinstance(self.widget.data, Table))
        self.assertEqual(len(self.panglao), len(self.widget.data))

    def test_organism_changed(self):
        """
        Test changing organisms in combination with changing the source.
        """
        simulate.combobox_activate_index(self.widget.controls.source_index, 0)
        simulate.combobox_activate_index(self.widget.controls.organism_index, 0)
        self.assertEqual("Panglao", self.widget.db_source_cb.currentText())
        self.assertEqual("Human", self.widget.group_cb.currentText())

        human_rows = self.widget.data.get_column_view("Organism")[0] == "Human"
        cell_types = self.widget.data.get_column_view("Cell Type")[0]

        model = self.widget.available_markers_view.model().sourceModel()
        self.assertEqual(sum(self.panglao.get_column_view("Organism")[0] == "Human"), len(model))
        self.assertEqual(len(np.unique(cell_types[human_rows])), len(model.rootItem.childItems))

        simulate.combobox_activate_index(self.widget.controls.source_index, 0)
        simulate.combobox_activate_index(self.widget.controls.organism_index, 1)
        self.assertEqual("Panglao", self.widget.db_source_cb.currentText())
        self.assertEqual("Mouse", self.widget.group_cb.currentText())

        model = self.widget.available_markers_view.model().sourceModel()
        self.assertEqual(sum(self.panglao.get_column_view("Organism")[0] == "Mouse"), len(model))
        self.assertEqual(len(np.unique(cell_types[~human_rows])), len(model.rootItem.childItems))

        simulate.combobox_activate_index(self.widget.controls.source_index, 1)
        simulate.combobox_activate_index(self.widget.controls.organism_index, 0)
        self.assertEqual("CellMarker", self.widget.db_source_cb.currentText())
        self.assertEqual("Human", self.widget.group_cb.currentText())

        human_rows = self.widget.data.get_column_view("Organism")[0] == "Human"
        cell_types = self.widget.data.get_column_view("Cell Type")[0]

        model = self.widget.available_markers_view.model().sourceModel()
        self.assertEqual(sum(self.cell_markers.get_column_view("Organism")[0] == "Human"), len(model))
        self.assertEqual(len(np.unique(cell_types[human_rows])), len(model.rootItem.childItems))

        simulate.combobox_activate_index(self.widget.controls.source_index, 1)
        simulate.combobox_activate_index(self.widget.controls.organism_index, 1)
        self.assertEqual("CellMarker", self.widget.db_source_cb.currentText())
        self.assertEqual("Mouse", self.widget.group_cb.currentText())

        model = self.widget.available_markers_view.model().sourceModel()
        self.assertEqual(sum(self.cell_markers.get_column_view("Organism")[0] == "Mouse"), len(model))
        self.assertEqual(len(np.unique(cell_types[~human_rows])), len(model.rootItem.childItems))

    def test_group_by_changed(self):
        """
        Test changing the group by parameter (Cell Type, Function)
        """
        # group by cell type
        simulate.combobox_activate_index(self.widget.controls.selected_root_attribute, 0)

        human_rows = self.widget.data.get_column_view("Organism")[0] == "Human"
        cell_types = np.unique(self.widget.data.get_column_view("Cell Type")[0][human_rows])
        model = self.widget.available_markers_view.model().sourceModel()
        self.assertEqual(len(cell_types), len(model.rootItem.childItems))

        # group by function
        simulate.combobox_activate_index(self.widget.controls.selected_root_attribute, 1)

        functions = np.unique(self.widget.data.get_column_view("Function")[0][human_rows])
        model = self.widget.available_markers_view.model().sourceModel()
        self.assertEqual(len(functions), len(model.rootItem.childItems))

    def test_move_elements(self):
        """
        Move elements between booth views and observe if numbers matches.
        This test just move the single element.
        """
        # select first three elements
        src_view = self.widget.available_markers_view
        dest_view = self.widget.selected_markers_view
        num_items_source_view = len(src_view.model().sourceModel())
        num_root_items = src_view.model().rowCount(QModelIndex())

        # select first element
        index = src_view.model().index(0, 0)
        src_view.selectionModel().select(index, QItemSelectionModel.Select | QItemSelectionModel.Rows)
        num_items = src_view.model().rowCount(index)  # number of items under selected item

        # move element to the TreeView with selected genes
        self.widget.move_button.click()
        self.assertEqual(len(dest_view.model().sourceModel()), num_items)
        output = self.get_output(self.widget.Outputs.genes)
        self.assertEqual(num_items, len(output))
        self.assertEqual(dest_view.model().rowCount(QModelIndex()), 1)  # one root element added
        index = dest_view.model().index(0, 0)
        self.assertEqual(dest_view.model().rowCount(index), num_items)

        self.assertEqual(src_view.model().rowCount(QModelIndex()), num_root_items - 1)  # one item is remove from view

        # move everything back
        dest_view.selectionModel().select(index, QItemSelectionModel.Select | QItemSelectionModel.Rows)
        self.widget.move_button.click()

        self.assertEqual(num_items_source_view, len(src_view.model().sourceModel()))
        self.assertEqual(src_view.model().rowCount(QModelIndex()), num_root_items)

    def test_move_more_elements(self):
        """
        Move elements between booth views and observe if numbers matches.
        This test keep moving more elements.
        """
        # select first three elements
        src_view = self.widget.available_markers_view
        dest_view = self.widget.selected_markers_view
        num_items_source_view = len(src_view.model().sourceModel())
        num_root_items = src_view.model().rowCount(QModelIndex())

        # select first element
        index = src_view.model().index(0, 0)
        index1 = src_view.model().index(1, 0)
        src_view.selectionModel().select(index, QItemSelectionModel.Select | QItemSelectionModel.Rows)
        src_view.selectionModel().select(index1, QItemSelectionModel.Select | QItemSelectionModel.Rows)
        num_items = src_view.model().rowCount(index)
        num_items1 = src_view.model().rowCount(index1)

        # move elements to the TreeView with selected genes
        self.widget.move_button.click()
        self.assertEqual(len(dest_view.model().sourceModel()), num_items + num_items1)
        output = self.get_output(self.widget.Outputs.genes)
        self.assertEqual(num_items + num_items1, len(output))
        self.assertEqual(dest_view.model().rowCount(QModelIndex()), 2)  # two root element added
        index = dest_view.model().index(0, 0)
        index1 = dest_view.model().index(1, 0)
        self.assertEqual(
            dest_view.model().rowCount(index) + dest_view.model().rowCount(index1), num_items + num_items1
        )

        self.assertEqual(src_view.model().rowCount(QModelIndex()), num_root_items - 2)

        # move everything back one by one
        dest_view.selectionModel().select(index, QItemSelectionModel.Select | QItemSelectionModel.Rows)
        self.widget.move_button.click()
        index1 = dest_view.model().index(0, 0)
        self.assertEqual(dest_view.model().rowCount(index1), num_items1)
        output = self.get_output(self.widget.Outputs.genes)
        self.assertEqual(num_items1, len(output))
        self.assertEqual(len(dest_view.model().sourceModel()), num_items1)
        self.assertEqual(src_view.model().rowCount(QModelIndex()), num_root_items - 1)

        # move the other one back
        dest_view.selectionModel().select(index1, QItemSelectionModel.Select | QItemSelectionModel.Rows)
        self.widget.move_button.click()
        self.assertEqual(num_items_source_view, len(src_view.model().sourceModel()))
        self.assertEqual(0, len(dest_view.model().sourceModel()))
        self.assertEqual(src_view.model().rowCount(QModelIndex()), num_root_items)
        self.assertEqual(dest_view.model().rowCount(QModelIndex()), 0)

    def test_move_genes(self):
        """
        Move elements between booth views and observe if numbers matches.
        In previous test we were moving just root elements. In this test we will move children elements.
        """
        src_view = self.widget.available_markers_view
        dest_view = self.widget.selected_markers_view
        num_items_source_view = len(src_view.model().sourceModel())
        num_root_items = src_view.model().rowCount(QModelIndex())

        # select first element
        parent_index = src_view.model().index(0, 0)
        index = src_view.model().index(0, 0, parent_index)
        index1 = src_view.model().index(1, 0, parent_index)

        src_view.selectionModel().select(index, QItemSelectionModel.Select | QItemSelectionModel.Rows)
        src_view.selectionModel().select(index1, QItemSelectionModel.Select | QItemSelectionModel.Rows)

        # move elements to the TreeView with selected genes
        self.widget.move_button.click()
        self.assertEqual(len(dest_view.model().sourceModel()), 2)
        output = self.get_output(self.widget.Outputs.genes)
        self.assertEqual(2, len(output))
        self.assertEqual(dest_view.model().rowCount(QModelIndex()), 1)  # one parent element added
        parent_index = dest_view.model().index(0, 0)
        self.assertEqual(dest_view.model().rowCount(parent_index), 2)
        self.assertEqual(src_view.model().rowCount(QModelIndex()), num_root_items)

        # move everything back one by one
        index = dest_view.model().index(0, 0, parent_index)
        dest_view.selectionModel().select(index, QItemSelectionModel.Select | QItemSelectionModel.Rows)
        self.widget.move_button.click()

        self.assertEqual(dest_view.model().rowCount(parent_index), 1)
        output = self.get_output(self.widget.Outputs.genes)
        self.assertEqual(1, len(output))
        self.assertEqual(len(dest_view.model().sourceModel()), 1)
        self.assertEqual(src_view.model().rowCount(QModelIndex()), num_root_items)

        # move the other one back
        index1 = dest_view.model().index(0, 0, parent_index)
        dest_view.selectionModel().select(index1, QItemSelectionModel.Select | QItemSelectionModel.Rows)
        self.widget.move_button.click()
        self.assertEqual(num_items_source_view, len(src_view.model().sourceModel()))
        self.assertEqual(src_view.model().rowCount(QModelIndex()), num_root_items)

    def test_description(self):
        """
        Test correctness of the gene description showing in the bottom of widget.
        """
        src_view = self.widget.available_markers_view
        dest_view = self.widget.selected_markers_view

        not_gene = "Select a gene to see information."
        first_gene_info = "Gene name: AKR1C3"

        # first test in the source view
        parent_index = src_view.model().index(0, 0)
        gene_index = src_view.model().index(0, 0, parent_index)
        gene_index1 = src_view.model().index(1, 0, parent_index)

        # no description for parent
        src_view.selectionModel().select(parent_index, QItemSelectionModel.Select | QItemSelectionModel.Rows)
        self.assertEqual(not_gene, self.widget.descriptionlabel.toPlainText())

        # no description for multiple genes
        src_view.selectionModel().select(gene_index, QItemSelectionModel.Select | QItemSelectionModel.Rows)
        src_view.selectionModel().select(gene_index1, QItemSelectionModel.Select | QItemSelectionModel.Rows)
        self.assertEqual(not_gene, self.widget.descriptionlabel.toPlainText())

        # description for first gene
        src_view.selectionModel().select(gene_index, QItemSelectionModel.ClearAndSelect | QItemSelectionModel.Rows)
        self.assertIn(first_gene_info, self.widget.descriptionlabel.toPlainText())

        # move elements
        src_view.selectionModel().select(gene_index, QItemSelectionModel.Select | QItemSelectionModel.Rows)
        src_view.selectionModel().select(gene_index1, QItemSelectionModel.Select | QItemSelectionModel.Rows)
        self.widget.move_button.click()

        # test in the dest view
        parent_index = dest_view.model().index(0, 0)
        gene_index = dest_view.model().index(0, 0, parent_index)
        gene_index1 = dest_view.model().index(1, 0, parent_index)

        # no description for parent
        dest_view.selectionModel().select(parent_index, QItemSelectionModel.Select | QItemSelectionModel.Rows)
        self.assertEqual(not_gene, self.widget.descriptionlabel.toPlainText())

        # no description for multiple genes
        dest_view.selectionModel().select(gene_index, QItemSelectionModel.Select | QItemSelectionModel.Rows)
        dest_view.selectionModel().select(gene_index1, QItemSelectionModel.Select | QItemSelectionModel.Rows)
        self.assertEqual(not_gene, self.widget.descriptionlabel.toPlainText())

        # description for first gene
        dest_view.selectionModel().select(gene_index, QItemSelectionModel.ClearAndSelect | QItemSelectionModel.Rows)
        self.assertIn(first_gene_info, self.widget.descriptionlabel.toPlainText())

    def test_extend_collapse_both_views(self):
        """
        Test whether the item extends in the other column when extended in one.
        """
        src_view = self.widget.available_markers_view
        dest_view = self.widget.selected_markers_view

        # I do not know how to simulate click so I just call the function that is called on click
        index = src_view.model().index(0, 0)
        src_view.expand_in_other_view(index, is_expanded=True)
        tree_node = src_view.model().sourceModel().node_from_index(src_view.model().mapToSource(index))
        self.assertTrue(tree_node.expanded)

        # move one child left to check simultaneous expand
        self.move_first_child(src_view)
        index_left = src_view.model().index(0, 0)
        index_right = dest_view.model().index(0, 0)
        tree_node_left = src_view.model().sourceModel().node_from_index(src_view.model().mapToSource(index_left))
        tree_node_right = dest_view.model().sourceModel().node_from_index(dest_view.model().mapToSource(index_right))

        src_view.expand_in_other_view(index_left, is_expanded=True)
        self.assertTrue(tree_node_left.expanded)
        self.assertTrue(tree_node_right.expanded)
        self.assertTrue(dest_view.isExpanded(index_right))

        # change back to not expanded
        src_view.expand_in_other_view(index_left, is_expanded=False)
        self.assertFalse(tree_node_left.expanded)
        self.assertFalse(tree_node_right.expanded)
        self.assertFalse(dest_view.isExpanded(index_right))

        dest_view.expand_in_other_view(index_right, is_expanded=True)
        self.assertTrue(tree_node_left.expanded)
        self.assertTrue(tree_node_right.expanded)
        self.assertTrue(src_view.isExpanded(index_left))

        # change back to not expanded
        dest_view.expand_in_other_view(index_right, is_expanded=False)
        self.assertFalse(tree_node_left.expanded)
        self.assertFalse(tree_node_right.expanded)
        self.assertFalse(dest_view.isExpanded(index_left))


if __name__ == '__main__':
    unittest.main()



import unittest

import numpy as np
from scipy.stats import multivariate_normal as mvn

from orangecontrib.bioinformatics.utils import statistics


class TestStatistics(unittest.TestCase):
    def test_hypergeometric(self):
        threshold = 1

        # Expression matrix (cells x genes)
        x = np.array(
            [
                [1.2, 0.38, 0.14, 0.5, 0.94, 0.55, 3.88, 0.49],
                [0.26, 1.05, 1.83, 0.35, 0.19, 2.62, 0.69, 0.79],
                [0.28, 0.91, 1.78, 0.38, 0.22, 3.25, 0.66, 0.61],
                [0.26, 1.02, 1.61, 0.35, 0.2, 2.67, 0.56, 0.62],
                [0.81, 1.45, 0.48, 0.85, 3.57, 0.63, 0.16, 0.52],
                [0.78, 2.04, 0.53, 0.83, 3.05, 0.71, 0.14, 0.63],
                [0.65, 1.84, 0.54, 0.98, 3.11, 0.58, 0.16, 0.64],
                [1.34, 0.43, 0.16, 0.57, 0.73, 0.55, 4.11, 0.4],
                [1.49, 0.41, 0.15, 0.6, 0.73, 0.44, 3.49, 0.48],
                [0.29, 0.87, 1.57, 0.36, 0.18, 2.67, 0.65, 0.65],
            ]
        )

        cluster = np.zeros((x.shape[0],), dtype=bool)
        cluster[:4] = True
        scores, pvalues = statistics.score_hypergeometric_test(
            a=x[cluster], b=x[np.logical_not(cluster),], threshold=threshold
        )
        self.assertIsNotNone(scores)

    def test_alternatives(self):
        """ Test implemented alternative hypotheses. """
        np.random.seed(42)
        n = 100
        a = mvn.rvs(mean=0, cov=1, size=n).reshape((n, 1))
        b = mvn.rvs(mean=1, cov=1, size=n).reshape((n, 1))
        c = mvn.rvs(mean=-1, cov=1, size=n).reshape((n, 1))

        for method in (statistics.score_t_test, statistics.score_hypergeometric_test, statistics.score_mann_whitney):

            _, pval_less = method(a, b, alternative=statistics.ALT_LESS)
            _, pval_two = method(a, b, alternative=statistics.ALT_TWO)
            _, pval_greater = method(a, b, alternative=statistics.ALT_GREATER)
            assert np.all(pval_less < pval_two < pval_greater)
            assert np.linalg.norm(pval_less + pval_greater - np.array([1.0])) < 1e-8

            _, pval_less = method(b, c, alternative=statistics.ALT_LESS)
            _, pval_two = method(b, c, alternative=statistics.ALT_TWO)
            _, pval_greater = method(b, c, alternative=statistics.ALT_GREATER)
            assert np.all(pval_greater < pval_two < pval_less)
            assert np.linalg.norm(pval_less + pval_greater - np.array([1.0])) < 1e-8

    def test_alternatives_2d(self):
        """ Test implemented alternative hypotheses. """
        np.random.seed(42)
        n = 100
        a1 = mvn.rvs(mean=0, cov=1, size=n).reshape((n, 1))
        a2 = mvn.rvs(mean=1, cov=1, size=n).reshape((n, 1))
        b1 = mvn.rvs(mean=2, cov=1, size=n).reshape((n, 1))
        b2 = mvn.rvs(mean=3, cov=1, size=n).reshape((n, 1))
        _a = np.hstack((a1, a2))
        _b = np.hstack((b1, b2))

        for method in (statistics.score_t_test, statistics.score_hypergeometric_test, statistics.score_mann_whitney):
            _, pval_less = method(_a, _b, alternative=statistics.ALT_LESS)
            _, pval_two = method(_a, _b, alternative=statistics.ALT_TWO)
            _, pval_greater = method(_a, _b, alternative=statistics.ALT_GREATER)
            assert np.all(np.logical_and(pval_less < pval_two, pval_two < pval_greater))
            assert np.linalg.norm(pval_less + pval_greater - np.array([1.0])) < 1e-8

    def test_identical(self):
        """ Test with identical values. """
        n = 100
        a = np.random.rand(n, 1)
        b = np.random.rand(n, 1)
        _a = np.hstack([a, a])
        _b = np.hstack([a, b])
        for alt in statistics.ALTERNATIVES:
            for method in (
                statistics.score_t_test,
                statistics.score_hypergeometric_test,
                statistics.score_mann_whitney,
            ):
                p, r = method(_a, _b, alternative=alt)
                assert np.isnan(p).sum() == 0
                assert np.isnan(r).sum() == 0


if __name__ == '__main__':
    unittest.main()



import os
import unittest


def suite(loader=None, pattern='test*.py'):
    test_dir = os.path.dirname(__file__)
    if loader is None:
        loader = unittest.TestLoader()
    if pattern is None:
        pattern = 'test*.py'
    all_tests = [loader.discover(test_dir, pattern)]

    return unittest.TestSuite(all_tests)


def load_tests(loader, tests, pattern):
    return suite(loader, pattern)


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



"""
NCBI GEO module


GEO datasets are hosted on datasets.biolab.si and are not part of
a bioinformatics knowledge base (we can't use bioinformatics.utils.serverfiles module).

"""
import os
import json

import numpy as np
from serverfiles import LocalFiles, ServerFiles

from Orange.data import Table, Domain, StringVariable, DiscreteVariable
from Orange.data import filter as table_filter
from Orange.misc.environ import data_dir

from orangecontrib.bioinformatics.widgets.utils.data import TableAnnotation

domain = 'geo'
_local_cache_path = os.path.join(data_dir(), domain)
_all_info_file = os.path.join(_local_cache_path, '__INFO__')
_server_url = 'http://download.biolab.si/datasets/geo/'
pubmed_url = 'http://www.ncbi.nlm.nih.gov/pubmed/{}'

server_files = ServerFiles(server=_server_url)
local_files = LocalFiles(_local_cache_path, serverfiles=server_files)


def is_cached(gds_id):
    return os.path.exists(os.path.join(_local_cache_path, gds_id + '.tab'))


def info_cache(f):
    """Store content of __INFO__ file locally."""

    def wrapper():
        if not os.path.isfile(_all_info_file) or os.path.getsize(_all_info_file) == 0:
            with open(_all_info_file, 'w') as fp:
                json.dump(list(server_files.allinfo().items()), fp)

        return f()

    return wrapper


@info_cache
def dataset_all_info():
    """
    Return a list of all GEO datasets from databases.biolab.si.
    """

    with open(_all_info_file, 'r') as fp:
        return json.load(fp)


def dataset_download(gds_id, samples=None, transpose=False, callback=None):
    file_name = '{}.tab'.format(gds_id)
    file_path = local_files.localpath_download(file_name, extract=True, callback=callback)

    table = Table(file_path)
    title = table.name
    gds_info = local_files.info(file_name)
    table_annotations = {TableAnnotation.tax_id: gds_info['taxid']}

    if callback:
        callback()

    if samples is not None:
        filters = [table_filter.FilterStringList(sample, sample_types) for sample, sample_types in samples.items()]
        table = table_filter.Values(filters)(table)

        column_values = []
        for meta_var in samples.keys():
            column_values.append(table.get_column_view(table.domain[meta_var])[0])

        class_values = list(map('|'.join, zip(*column_values)))

        _class_values = list(set(class_values))
        map_class_values = {value: key for (key, value) in enumerate(_class_values)}
        class_var = DiscreteVariable(name='class', values=_class_values)
        _domain = Domain(table.domain.attributes, table.domain.class_vars + (class_var,), table.domain.metas)

        table = table.transform(_domain)
        with table.unlocked(table.Y):
            col, _ = table.get_column_view(class_var)
            col[:] = [map_class_values[class_val] for class_val in class_values]

    if transpose:
        table = Table.transpose(table, feature_names_column='sample_id', meta_attr_name='genes')

        # When transposing a table, variable.attributes get picked up as numerical values instead of strings.
        # We need to convert from Continuous to StringVariable
        _genes = [
            [str(int(gene)) if not np.isnan(gene) else '?']
            for gene in table.get_column_view('Entrez ID')[0].astype(np.float64)
        ]
        new_var = StringVariable('Entrez ID')
        metas = [var for var in table.domain.metas if var.name != 'Entrez ID'] + [new_var]
        new_domain = Domain(table.domain.attributes, table.domain.class_vars, metas)
        table = table.transform(new_domain)
        with table.unlocked(table.metas):
            table[:, new_var] = _genes

        # table name is lost after transpose
        table.name = title

        table_annotations[TableAnnotation.gene_as_attr_name] = not gds_info[TableAnnotation.gene_as_attr_name]
        table_annotations[TableAnnotation.gene_id_column] = gds_info[TableAnnotation.gene_id_attribute]
    else:
        table_annotations[TableAnnotation.gene_as_attr_name] = gds_info[TableAnnotation.gene_as_attr_name]
        table_annotations[TableAnnotation.gene_id_attribute] = gds_info[TableAnnotation.gene_id_attribute]

    if callback:
        callback()

    table.attributes = table_annotations
    return table



""" Resolwe module """
from .utils import GENAPI_CACHE, CACHE_BACKEND, RESOLWEAPI_CACHE
from .genapi import GenAPI
from .resapi import ResolweAPI

RESOLWE_PLATFORM = 'resolwe'
GENESIS_PLATFORM = 'genesis'


class ResolweAuthError(Exception):
    """A login error occurred."""


class ResolweServerTypeError(Exception):
    """Unknown server type"""


def connect(username=None, password=None, url=None, server_type='resolwe'):
    """Connect to Resolwe server

    :param username:
    :type username: str

    :param password:
    :type password: str

    :param url:
    :type url: str

    :param server_type: genesis or resolwe
    :type server_type: str

    :return: Instance of GenAPI or ResolweAPI
    """

    if server_type == RESOLWE_PLATFORM:
        _api = ResolweAPI
    elif server_type == GENESIS_PLATFORM:
        _api = GenAPI
    else:
        raise ResolweServerTypeError(f'Unknown server type {server_type}')

    try:
        return _api(username, password, url)
    except Exception as e:
        raise ResolweAuthError(e.args[0]) from e



"""
Wrapper classes for db entry fields to support pythonic
interface.

"""


class DBField(object):
    """
    Base DBGET entry field
    """

    __SLOTS__ = ["text"]

    def __init__(self, text):
        self.text = text

    def _convert(self):
        """
        Convert the contents into python representation using builtin types.
        """
        return self.text.rstrip("\n")


class DBSimpleField(DBField):
    """
    Simple field (with no subsections).
    """

    __SLOTS__ = ["text"]
    # TITLE must be set in subclasses or object instances
    TITLE = None

    def __str__(self):
        return self.format()

    def format(self, section_indent=12, subsection_indent=0):
        fmt = (" " * subsection_indent) + "%-" + str(section_indent - subsection_indent) + "s%s"
        text = self._indent(self.text, section_indent)
        text = fmt % (self.TITLE, text)
        return text

    def _indent(self, text, section_indent=12):
        indent_str = "\n" + " " * section_indent
        nl_count = text.count("\n")
        return text.replace("\n", indent_str, nl_count - 1)


class DBEntryField(DBSimpleField):
    """
    ENTRY field (all entries start with this field)
    """

    __SLOTS__ = ["text"]
    TITLE = "ENTRY"


class DBNameField(DBSimpleField):
    __SLOTS__ = ["text"]
    TITLE = "NAME"


class DBDefinitionField(DBSimpleField):
    __SLOTS__ = ["text"]
    TITLE = "DEFINITION"


class DBFieldWithSubsections(DBSimpleField):
    """
    A field with subsections (for instance REFERENCE in genome)
    """

    __SLOTS__ = ["text", "subsections"]
    TITLE = None
    SUBSECTIONS = None

    def __init__(self, text, subsections=None):
        self.text = text
        self.subsections = subsections or []

    def format(self, section_indent=12, subsection_indent=2):
        text = DBSimpleField.format(self, section_indent, subsection_indent=0)
        subsections = [sub.format(section_indent, subsection_indent) for sub in self.subsections]
        return "".join([text] + subsections)

    def _convert(self):
        my = DBSimpleField._convert(self)
        subs = [(s.TITLE.lower(), s._convert()) for s in self.subsections]
        return (my, subs)


class DBTaxonomyField(DBFieldWithSubsections):
    __SLOTS__ = ["text", "subsections"]
    TITLE = "TAXONOMY"
    SUBSECTIONS = ["LINEAGE"]

    @property
    def taxid(self):
        return DBSimpleField._convert(self).split(":")[1]


class DBDataSourceField(DBSimpleField):
    __SLOTS__ = ["text"]
    TITLE = "DATA_SOURCE"


class DBReference(DBFieldWithSubsections):
    __SLOTS__ = ["text", "subsections"]
    TITLE = "REFERENCE"
    SUBSECTIONS = ["AUTHORS", "TITLE", "JOURNAL"]

    @property
    def authors(self):
        return self.subsections[0]

    @property
    def title(self):
        return self.subsections[1]

    @property
    def journal(self):
        return self.subsections[2]


class DBDBLinks(DBSimpleField):
    __SLOTS__ = ["text"]
    TITLE = "DBLINKS"

    @property
    def links(self):
        return [tuple(s.split(": ", 1)) for s in self.text.splitlines()]

    def _convert(self):
        # Some dblinks can span multiple lines but are always 'indented'
        links = DBSimpleField._convert(self).replace("\n ", "").splitlines()
        links = [tuple(link.split(": ", 1)) for link in links]
        links = [(key, [v for v in values.split(" ") if v]) for key, values in links]
        return dict(links)


class DBPathway(DBSimpleField):
    __SLOTS__ = ["text"]
    TITLE = "PATHWAY"

    @property
    def pathways(self):
        return self._convert()

    def _convert(self):
        text = DBSimpleField._convert(self)
        return [line.split(" ", 1)[0] for line in text.splitlines()]


class DBAASeq(DBSimpleField):
    __SLOTS__ = ["text"]
    TITLE = "AASEQ"

    @property
    def sequence(self):
        return self.split("\n", 1)[1].replace("\n", "")

    @property
    def sequence_lenght(self):
        return int(self.text.split("\n", 1)[0])

    def _convert(self):
        text = DBSimpleField._convert(self)
        count, seq = text.split("\n", 1)
        return seq.replace("\n", "")


class DBNTSeq(DBSimpleField):
    __SLOTS__ = ["text"]
    TITLE = "NTSEQ"

    @property
    def sequence(self):
        return self.split("\n", 1)[1].replace("\n", "")

    @property
    def sequence_lenght(self):
        return int(self.text.split("\n", 1)[0])

    def _convert(self):
        text = DBSimpleField._convert(self)
        count, seq = text.split("\n", 1)
        return seq.replace("\n", "")


class DBPathwayMapField(DBSimpleField):
    __SLOTS__ = ["text"]
    TITLE = "PATHWAY_MAP"

    def kgml_url(self):
        return "http://www.genome.jp/kegg-bin/download?entry={0}&format=kgml".format(self.pathway_id)

    @property
    def pathway_id(self):
        return self.text.split(" ", 1)[0]


class DBGeneField(DBSimpleField):
    __SLOTS__ = ["text"]
    TITLE = "GENE"

    def _convert(self):
        text = DBSimpleField._convert(self)
        lines = text.splitlines()
        return [line.split(" ", 1)[0] for line in lines]


class DBEnzymeField(DBSimpleField):
    __SLOTS__ = ["text"]
    TITLE = "ENZYME"

    def _convert(self):
        text = DBSimpleField._convert(self)
        lines = text.splitlines()
        return lines


class DBCompoundField(DBSimpleField):
    __SLOTS__ = ["text"]
    TITLE = "COMPOUND"

    def _convert(self):
        text = DBSimpleField._convert(self)
        lines = text.splitlines()
        return [line.split(" ", 1)[0] for line in lines]



import random
import unittest
from datetime import datetime, timezone
from unittest.mock import call, patch

from AnyQt.QtCore import pyqtSignal

from Orange.widgets.widget import OWWidget
from Orange.widgets.settings import SettingProvider
from Orange.widgets.tests.base import WidgetTest
from Orange.widgets.credentials import CredentialManager

from orangecontrib.bioinformatics.resolwe import resapi
from orangecontrib.bioinformatics.widgets.components.resolwe import SignIn
from orangecontrib.bioinformatics.widgets.OWGenialisExpressions import (
    SortBy,
    ItemsPerPage,
    PaginationComponent,
    FilterByDateModified,
    OWGenialisExpressions,
    CollapsibleFilterComponent,
)


class MockWidget(OWWidget):
    name = 'MockWidget'
    want_main_area = False

    filter_component = SettingProvider(CollapsibleFilterComponent)
    pagination_component = SettingProvider(PaginationComponent)

    pagination_availability = pyqtSignal(bool, bool)

    @patch(
        'orangecontrib.bioinformatics.widgets.components.resolwe.get_credential_manager',
        return_value=CredentialManager('resolwe_credentials_test'),
    )
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.filter_component = CollapsibleFilterComponent(self, self.controlArea)
        self.pagination_component = PaginationComponent(self, self.controlArea)
        self.sign_in_dialog = SignIn(self)


def collection_generator(num_of_collections):
    return [
        {
            'id': int(i),
            'name': f'Collection name {i}',
            'contributor': {'first_name': f'Foo_{i}', 'last_name': f'Bar_{i}', 'username': f'FooBar_{i}'},
            'description': f'Test description {1}',
            'data_count': random.randint(0, 10),
            'entity_count': random.randint(0, 10),
            'tags': ['community:expressions'],
            'created': datetime.strftime(datetime.now().replace(tzinfo=timezone.utc), '%Y-%m-%dT%H:%M:%S.%f%z'),
            'modified': datetime.strftime(datetime.now().replace(tzinfo=timezone.utc), '%Y-%m-%dT%H:%M:%S.%f%z'),
        }
        for i in range(num_of_collections)
    ]


class TestOWGenialisExpressions(WidgetTest):
    @patch('orangecontrib.bioinformatics.resolwe.ResolweAPI', spec=True)
    @patch('orangecontrib.bioinformatics.widgets.OWGenialisExpressions.OWGenialisExpressions.sign_in')
    def test_signin_popup(self, mock_sign_in, _):
        self.widget = self.create_widget(OWGenialisExpressions)
        self.widget.sign_in_btn.click()
        mock_sign_in.assert_has_calls([call(silent=True), call(False)])

    @patch('orangecontrib.bioinformatics.resolwe.ResolweAPI', spec=True)
    def test_widget_initialization(self, mock_resolwe_api):
        count = 2
        mock_resolwe_api.return_value.get_collections.return_value = {
            'count': count,
            'next': None,
            'previous': None,
            'results': collection_generator(count),
        }
        self.widget = self.create_widget(OWGenialisExpressions)

        # Check if calls are made with defaults
        offset = 0
        limit = ItemsPerPage.values()[ItemsPerPage.min]
        ordering = SortBy.values()[SortBy.newest_first]
        mock_resolwe_api.assert_called_once_with(None, None, resapi.DEFAULT_URL)
        mock_resolwe_api.return_value.get_collections.assert_called_once_with(
            limit=limit, offset=offset, ordering=ordering
        )
        mock_resolwe_api.return_value.get_species.assert_called_once_with(list(range(count)))

        self.assertEqual(count, self.widget.model.rowCount())

    @patch('orangecontrib.bioinformatics.resolwe.ResolweAPI', spec=True)
    def test_sign_out(self, mock_resolwe_api):
        # collections before/after sign out
        count_before = 10
        count_after = 2

        mock_resolwe_api.return_value.get_collections.return_value = {
            'count': count_before,
            'next': None,
            'previous': None,
            'results': collection_generator(count_before),
        }
        self.widget = self.create_widget(OWGenialisExpressions)
        self.assertEqual(count_before, self.widget.model.rowCount())

        # check if state resets after sign out
        mock_resolwe_api.return_value.get_collections.return_value = {
            'count': count_after,
            'next': None,
            'previous': None,
            'results': collection_generator(count_after),
        }
        self.widget.sign_out_btn.click()
        self.assertEqual(count_after, self.widget.model.rowCount())

    @patch('orangecontrib.bioinformatics.resolwe.ResolweAPI', spec=True)
    def test_query_params(self, mock_resolwe_api):
        self.widget = self.create_widget(OWGenialisExpressions)

        full_text = 'test_fts'
        self.widget.filter_component.filter_full_text.setText(full_text)

        col_name = 'test_name'
        self.widget.filter_component.filter_name.setText(col_name)

        col_owner = 'test_owner'
        self.widget.filter_component.filter_owner.setText(col_owner)

        col_contrib = 'test_contrib'
        self.widget.filter_component.filter_contrib.setText(col_contrib)

        self.widget.filter_component.filter_by_modified = FilterByDateModified.past_hour

        self.widget.on_filter_changed()
        _, query_params = mock_resolwe_api.return_value.get_collections.call_args
        self.assertIn('limit', query_params)
        self.assertIn('offset', query_params)
        self.assertIn('ordering', query_params)
        self.assertIn('ordering', query_params)
        self.assertIn('text', query_params)
        self.assertIn('name__icontains', query_params)
        self.assertIn('contributor_name', query_params)
        self.assertIn('owners_name', query_params)
        self.assertIn('modified__gte', query_params)
        self.assertIn(full_text, query_params.values())
        self.assertIn(col_name, query_params.values())
        self.assertIn(col_owner, query_params.values())
        self.assertIn(col_contrib, query_params.values())


if __name__ == "__main__":
    unittest.main()



""" GO Enrichment Analysis """
import gc
import sys
import math
import logging
import operator
import itertools
import webbrowser
from types import SimpleNamespace
from typing import Dict, List, Tuple
from functools import reduce
from collections import defaultdict
from concurrent.futures import Future

import numpy
from requests.exceptions import ConnectTimeout, ConnectionError, RequestException

from AnyQt.QtGui import QBrush
from AnyQt.QtCore import Qt, QSize, QTimer, Signal, QThread
from AnyQt.QtCore import pyqtSlot as Slot
from AnyQt.QtWidgets import (
    QMenu,
    QLabel,
    QDialog,
    QCheckBox,
    QSplitter,
    QTreeView,
    QTreeWidget,
    QVBoxLayout,
    QItemDelegate,
    QTreeWidgetItem,
)

import Orange.data
from Orange.widgets import gui, widget, settings
from Orange.widgets.utils.concurrent import Task, FutureWatcher, ThreadExecutor, methodinvoke

from orangecontrib.bioinformatics import go
from orangecontrib.bioinformatics.ncbi import gene, taxonomy
from orangecontrib.bioinformatics.utils import statistics, serverfiles
from orangecontrib.bioinformatics.go.config import DOMAIN, FILENAME_ONTOLOGY, FILENAME_ANNOTATION
from orangecontrib.bioinformatics.widgets.utils.data import TableAnnotation, check_table_annotation


class EnsureDownloaded(Task):
    advance = Signal()
    progress = Signal(float)

    def __init__(self, files_list, parent=None):
        Task.__init__(self, parent)
        self.files_list = files_list

    def run(self):
        nfiles = len(self.files_list)
        count = 100 * nfiles
        counter = itertools.count()

        def advance():
            self.advance.emit()
            self.progress.emit(100.0 * next(counter) / count)

        for domain, filename in self.files_list:
            ensure_downloaded(domain, filename, advance=advance)


def ensure_downloaded(domain, filename, advance=None):
    serverfiles.localpath_download(domain, filename, callback=advance)


def isstring(var):
    return isinstance(var, Orange.data.StringVariable)


class TreeNode(object):
    def __init__(self, value, children):
        self.value = value
        self.children = children


class GOTreeWidget(QTreeWidget):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self._currMenu = QMenu()
        self._currAction = self._currMenu.addAction("View term on AmiGO website")

    def contextMenuEvent(self, event):
        super().contextMenuEvent(event)

        def browser_action(_term):
            if isinstance(_term, go.Term):
                _term = _term.id
            webbrowser.open(f'http://amigo.geneontology.org/amigo/term/{_term}')

        term = self.itemAt(event.pos()).term
        self._currAction.triggered.connect(lambda: browser_action(term))
        self._currMenu.popup(event.globalPos())


class State:
    #: Ready to run
    Ready = 1
    #: Downloading datasets needed to run.
    Downloading = 2
    #: Running the enrichment task
    Running = 4
    #: The current executing task is stale. Need to reschedule another update
    #: once this one completes.
    #: Only applicable with Downloading and Running
    Stale = 8


class OWGOBrowser(widget.OWWidget):
    name = "GO Browser"
    description = "Enrichment analysis for Gene Ontology terms."
    icon = "../widgets/icons/OWGOBrowser.svg"
    priority = 60

    inputs = [
        ("Cluster Data", Orange.data.Table, "set_dataset", widget.Single + widget.Default),
        ("Reference Data", Orange.data.Table, "set_reference_dataset"),
    ]

    outputs = [("Data on Selected Genes", Orange.data.Table), ("Enrichment Report", Orange.data.Table)]

    settingsHandler = settings.DomainContextHandler()

    gene_attr_index = settings.ContextSetting(0)
    use_attr_names = settings.ContextSetting(False)
    use_reference_dataset = settings.Setting(False)
    aspect_index = settings.Setting(0)
    use_evidence_type = settings.Setting({et: True for et in go.evidence_types_ordered})
    filter_by_num_of_instances = settings.Setting(False)
    min_num_of_instances = settings.Setting(1)
    filter_by_p_value = settings.Setting(True)
    max_p_value = settings.Setting(0.2)
    filter_by_p_value_nofdr = settings.Setting(False)
    max_p_value_no_fdr = settings.Setting(0.01)
    prob_func = settings.Setting(0)
    selection_direct_annotation = settings.Setting(0)
    selection_disjoint = settings.Setting(0)

    class Error(widget.OWWidget.Error):
        serverfiles_unavailable = widget.Msg(
            'Can not locate annotation files, ' 'please check your connection and try again.'
        )

    def __init__(self, parent=None):
        super().__init__(self, parent)

        self.input_data = None
        self.gene_info = None
        self.ref_data = None
        self.ontology = None
        self.annotations = None
        self.loaded_annotation_code = None
        self.treeStructRootKey = None
        self.probFunctions = [statistics.Binomial(), statistics.Hypergeometric()]
        self.selectedTerms = []

        self.selectionChanging = 0
        self.__state = State.Ready
        self.__scheduletimer = QTimer(self, singleShot=True)
        self.__scheduletimer.timeout.connect(self.__update)

        #############
        # GUI
        #############
        self.tabs = gui.tabWidget(self.controlArea)
        # Input tab
        self.inputTab = gui.createTabPage(self.tabs, "Input")
        box = gui.widgetBox(self.inputTab, "Info")
        self.infoLabel = gui.widgetLabel(box, "No data on input\n")

        gui.button(
            box,
            self,
            "Ontology/Annotation Info",
            callback=self.show_info,
            tooltip="Show information on loaded ontology and annotations",
        )

        self.referenceRadioBox = gui.radioButtonsInBox(
            self.inputTab,
            self,
            "use_reference_dataset",
            ["Entire genome", "Reference set (input)"],
            tooltips=["Use entire genome for reference", "Use genes from Referece Examples input signal as reference"],
            box="Reference",
            callback=self.__invalidate,
        )

        self.referenceRadioBox.buttons[1].setDisabled(True)
        gui.radioButtonsInBox(
            self.inputTab,
            self,
            "aspect_index",
            ["Biological process", "Cellular component", "Molecular function"],
            box="Aspect",
            callback=self.__invalidate,
        )

        # Filter tab
        self.filterTab = gui.createTabPage(self.tabs, "Filter")
        box = gui.widgetBox(self.filterTab, "Filter GO Term Nodes")
        gui.checkBox(
            box,
            self,
            "filter_by_num_of_instances",
            "Genes",
            callback=self.filter_and_display_graph,
            tooltip="Filter by number of input genes mapped to a term",
        )
        ibox = gui.indentedBox(box)
        gui.spin(
            ibox,
            self,
            'min_num_of_instances',
            1,
            100,
            step=1,
            label='#:',
            labelWidth=15,
            callback=self.filter_and_display_graph,
            callbackOnReturn=True,
            tooltip="Min. number of input genes mapped to a term",
        )

        gui.checkBox(
            box,
            self,
            "filter_by_p_value_nofdr",
            "p-value",
            callback=self.filter_and_display_graph,
            tooltip="Filter by term p-value",
        )

        gui.doubleSpin(
            gui.indentedBox(box),
            self,
            'max_p_value_no_fdr',
            1e-8,
            1,
            step=1e-8,
            label='p:',
            labelWidth=15,
            callback=self.filter_and_display_graph,
            callbackOnReturn=True,
            tooltip="Max term p-value",
        )

        # use filter_by_p_value for FDR, as it was the default in prior versions
        gui.checkBox(
            box, self, "filter_by_p_value", "FDR", callback=self.filter_and_display_graph, tooltip="Filter by term FDR"
        )
        gui.doubleSpin(
            gui.indentedBox(box),
            self,
            'max_p_value',
            1e-8,
            1,
            step=1e-8,
            label='p:',
            labelWidth=15,
            callback=self.filter_and_display_graph,
            callbackOnReturn=True,
            tooltip="Max term p-value",
        )

        box = gui.widgetBox(box, "Significance test")

        gui.radioButtonsInBox(
            box,
            self,
            "prob_func",
            ["Binomial", "Hypergeometric"],
            tooltips=["Use binomial distribution test", "Use hypergeometric distribution test"],
            callback=self.__invalidate,
        )  # TODO: only update the p values
        box = gui.widgetBox(self.filterTab, "Evidence codes in annotation", addSpace=True)
        self.evidenceCheckBoxDict = {}
        for etype in go.evidence_types_ordered:
            ecb = QCheckBox(etype, toolTip=go.evidence_types[etype], checked=self.use_evidence_type[etype])
            ecb.toggled.connect(self.__on_evidence_changed)
            box.layout().addWidget(ecb)
            self.evidenceCheckBoxDict[etype] = ecb

        # Select tab
        self.selectTab = gui.createTabPage(self.tabs, "Select")
        box = gui.radioButtonsInBox(
            self.selectTab,
            self,
            "selection_direct_annotation",
            ["Directly or Indirectly", "Directly"],
            box="Annotated genes",
            callback=self.example_selection,
        )

        box = gui.widgetBox(self.selectTab, "Output", addSpace=True)
        gui.radioButtonsInBox(
            box,
            self,
            "selection_disjoint",
            btnLabels=["All selected genes", "Term-specific genes", "Common term genes"],
            tooltips=[
                "Outputs genes annotated to all selected GO terms",
                "Outputs genes that appear in only one of selected GO terms",
                "Outputs genes common to all selected GO terms",
            ],
            callback=self.example_selection,
        )

        # ListView for DAG, and table for significant GOIDs
        self.DAGcolumns = ['GO term', 'Cluster', 'Reference', 'p-value', 'FDR', 'Genes', 'Enrichment']

        self.splitter = QSplitter(Qt.Vertical, self.mainArea)
        self.mainArea.layout().addWidget(self.splitter)

        # list view
        self.listView = GOTreeWidget(self.splitter)
        self.listView.setSelectionMode(QTreeView.ExtendedSelection)
        self.listView.setAllColumnsShowFocus(1)
        self.listView.setColumnCount(len(self.DAGcolumns))
        self.listView.setHeaderLabels(self.DAGcolumns)

        self.listView.header().setSectionsClickable(True)
        self.listView.header().setSortIndicatorShown(True)
        self.listView.header().setSortIndicator(self.DAGcolumns.index('p-value'), Qt.AscendingOrder)
        self.listView.setSortingEnabled(True)
        self.listView.setItemDelegateForColumn(6, EnrichmentColumnItemDelegate(self))
        self.listView.setRootIsDecorated(True)

        self.listView.itemSelectionChanged.connect(self.view_selection_changed)

        # table of significant GO terms
        self.sigTerms = QTreeWidget(self.splitter)
        self.sigTerms.setColumnCount(len(self.DAGcolumns))
        self.sigTerms.setHeaderLabels(self.DAGcolumns)
        self.sigTerms.setSortingEnabled(True)
        self.sigTerms.setSelectionMode(QTreeView.ExtendedSelection)
        self.sigTerms.header().setSortIndicator(self.DAGcolumns.index('p-value'), Qt.AscendingOrder)
        self.sigTerms.setItemDelegateForColumn(6, EnrichmentColumnItemDelegate(self))

        self.sigTerms.itemSelectionChanged.connect(self.table_selection_changed)

        self.sigTableTermsSorted = []
        self.graph = {}
        self.originalGraph = None

        self.inputTab.layout().addStretch(1)
        self.filterTab.layout().addStretch(1)
        self.selectTab.layout().addStretch(1)

        class AnnotationSlot(SimpleNamespace):
            taxid = ...  # type: str
            name = ...  # type: str
            filename = ...  # type:str

            @staticmethod
            def parse_tax_id(f_name):
                return f_name.split('.')[0]

        try:
            remote_files = serverfiles.ServerFiles().listfiles(DOMAIN)
        except (ConnectTimeout, RequestException, ConnectionError):
            # TODO: Warn user about failed connection to the remote server
            remote_files = []

        self.available_annotations = [
            AnnotationSlot(
                taxid=AnnotationSlot.parse_tax_id(annotation_file),
                name=taxonomy.common_taxid_to_name(AnnotationSlot.parse_tax_id(annotation_file)),
                filename=FILENAME_ANNOTATION.format(AnnotationSlot.parse_tax_id(annotation_file)),
            )
            for _, annotation_file in set(remote_files + serverfiles.listfiles(DOMAIN))
            if annotation_file != FILENAME_ONTOLOGY
        ]
        self._executor = ThreadExecutor()

    def sizeHint(self):
        return QSize(1000, 700)

    def __on_evidence_changed(self):
        for etype, cb in self.evidenceCheckBoxDict.items():
            self.use_evidence_type[etype] = cb.isChecked()
        self.__invalidate()

    def clear(self):
        self.infoLabel.setText("No data on input\n")
        self.warning(0)
        self.warning(1)
        self.clear_graph()

        self.send("Data on Selected Genes", None)
        self.send("Enrichment Report", None)

    @check_table_annotation
    def set_dataset(self, data=None):
        self.closeContext()
        self.clear()
        self.Error.clear()
        # print(data)
        if data:
            self.input_data = data
            self.tax_id = data.attributes[TableAnnotation.tax_id]
            self.use_attr_names = data.attributes[TableAnnotation.gene_as_attr_name]
            if self.use_attr_names:
                self.gene_id_attribute = data.attributes[TableAnnotation.gene_id_attribute]
            else:
                self.gene_id_column = data.attributes[TableAnnotation.gene_id_column]
            self.annotation_index = None

            _c2i = {a.taxid: i for i, a in enumerate(self.available_annotations)}
            try:
                self.annotation_index = _c2i[self.tax_id]
            except KeyError:
                self.Error.serverfiles_unavailable()
                # raise ValueError('Taxonomy {} not supported.'.format(self.tax_id))
                return

            self.gene_info = gene.GeneInfo(self.tax_id)

            self.__invalidate()

    @check_table_annotation
    def set_reference_dataset(self, data=None):
        self.Error.clear()
        if data:
            self.ref_data = data
            self.ref_tax_id = data.attributes[TableAnnotation.tax_id]
            self.ref_use_attr_names = data.attributes[TableAnnotation.gene_as_attr_name]
            if self.ref_use_attr_names:
                self.ref_gene_id_attribute = data.attributes[TableAnnotation.gene_id_attribute]
            else:
                self.ref_gene_id_column = data.attributes[TableAnnotation.gene_id_column]

        self.referenceRadioBox.buttons[1].setDisabled(not bool(data))
        self.referenceRadioBox.buttons[1].setText("Reference set")
        if self.input_data is not None and self.use_reference_dataset:
            self.use_reference_dataset = 0 if not data else 1
            self.__invalidate()

    @Slot()
    def __invalidate(self):
        # Invalidate the current results or pending task and schedule an
        # update.
        self.__scheduletimer.start()
        if self.__state != State.Ready:
            self.__state |= State.Stale

        self.set_graph({})
        self.ref_genes = None
        self.input_genes = None

    def __invalidate_annotations(self):
        self.annotations = None
        self.loaded_annotation_code = None
        if self.input_data:
            self.infoLabel.setText("...\n")
        self.__invalidate()

    @Slot()
    def __update(self):
        self.__scheduletimer.stop()
        if self.input_data is None:
            return

        if self.__state & State.Running:
            self.__state |= State.Stale
        elif self.__state & State.Downloading:
            self.__state |= State.Stale
        elif self.__state & State.Ready:
            if self.__ensure_data():
                self.load()
                self.enrichment()
            else:
                assert self.__state & State.Downloading
                assert self.isBlocking()

    def __get_ref_genes(self):
        self.ref_genes = []

        if self.ref_use_attr_names:
            for variable in self.input_data.domain.attributes:
                self.ref_genes.append(str(variable.attributes.get(self.ref_gene_id_attribute, '?')))
        else:
            genes, _ = self.ref_data.get_column_view(self.ref_gene_id_column)
            self.ref_genes = [str(g) for g in genes]

    def __get_input_genes(self):
        self.input_genes = []

        if self.use_attr_names:
            for variable in self.input_data.domain.attributes:
                self.input_genes.append(str(variable.attributes.get(self.gene_id_attribute, '?')))
        else:
            genes, _ = self.input_data.get_column_view(self.gene_id_column)
            self.input_genes = [str(g) for g in genes]

    def filter_annotated_genes(self, genes):
        matchedgenes = self.annotations.get_gene_names_translator(genes).values()
        return matchedgenes, [gene for gene in genes if gene not in matchedgenes]

    def __start_download(self, files_list):
        # type: (List[Tuple[str, str]]) -> None
        task = EnsureDownloaded(files_list)
        task.progress.connect(self._progress_bar_set)

        f = self._executor.submit(task)
        fw = FutureWatcher(f, self)
        fw.finished.connect(self.__download_finish)
        fw.finished.connect(fw.deleteLater)
        fw.resultReady.connect(self.__invalidate)

        self.progressBarInit(processEvents=None)
        self.setBlocking(True)
        self.setStatusMessage("Downloading")
        self.__state = State.Downloading

    @Slot(Future)
    def __download_finish(self, result):
        # type: (Future[None]) -> None
        assert QThread.currentThread() is self.thread()
        assert result.done()
        self.setBlocking(False)
        self.setStatusMessage("")
        self.progressBarFinished(processEvents=False)
        try:
            result.result()
        except ConnectTimeout:
            logging.getLogger(__name__).error("Error:")
            self.error(
                2,
                "Internet connection error, unable to load data. "
                + "Check connection and create a new GO Browser widget.",
            )
        except RequestException as err:
            logging.getLogger(__name__).error("Error:")
            self.error(2, "Internet error:\n" + str(err))
        except BaseException as err:
            logging.getLogger(__name__).error("Error:")
            self.error(2, "Error:\n" + str(err))
            raise
        else:
            self.error(2)
        finally:
            self.__state = State.Ready

    def __ensure_data(self):
        # Ensure that all required database (ontology and annotations for
        # the current selected organism are present. If not start a download in
        # the background. Return True if all dbs are present and false
        # otherwise
        assert self.__state == State.Ready
        annotation = self.available_annotations[self.annotation_index]
        go_files = [fname for domain, fname in serverfiles.listfiles(DOMAIN)]
        files = []

        if annotation.filename not in go_files:
            files.append(("go", annotation.filename))

        if FILENAME_ONTOLOGY not in go_files:
            files.append((DOMAIN, FILENAME_ONTOLOGY))
        if files:
            self.__start_download(files)
            assert self.__state == State.Downloading
            return False
        else:
            return True

    def load(self):
        a = self.available_annotations[self.annotation_index]

        if self.ontology is None:
            self.ontology = go.Ontology()

        if a.taxid != self.loaded_annotation_code:
            self.annotations = None
            gc.collect()  # Force run garbage collection
            self.annotations = go.Annotations(a.taxid)
            self.loaded_annotation_code = a.taxid
            count = defaultdict(int)
            gene_sets = defaultdict(set)

            for anno in self.annotations.annotations:
                count[anno.evidence] += 1
                gene_sets[anno.evidence].add(anno.gene_id)
            for etype in go.evidence_types_ordered:
                ecb = self.evidenceCheckBoxDict[etype]
                ecb.setEnabled(bool(count[etype]))
                ecb.setText(etype + ": %i annots(%i genes)" % (count[etype], len(gene_sets[etype])))

    def enrichment(self):
        assert self.input_data is not None
        assert self.__state == State.Ready

        if not self.annotations.ontology:
            self.annotations.ontology = self.ontology

        self.error(1)
        self.warning([0, 1])

        self.__get_input_genes()
        self.input_genes = set(self.input_genes)
        self.known_input_genes = self.annotations.get_genes_with_known_annotation(self.input_genes)

        # self.clusterGenes = clusterGenes = self.annotations.map_to_ncbi_id(self.input_genes).values()

        self.infoLabel.setText(
            "%i unique genes on input\n%i (%.1f%%) genes with known annotations"
            % (
                len(self.input_genes),
                len(self.known_input_genes),
                100.0 * len(self.known_input_genes) / len(self.input_genes) if len(self.input_genes) else 0.0,
            )
        )

        if not self.use_reference_dataset or self.ref_data is None:
            self.information(2)
            self.information(1)
            self.ref_genes = set(self.gene_info.keys())

        elif self.ref_data is not None:
            self.__get_ref_genes()
            self.ref_genes = set(self.ref_genes)

            ref_count = len(self.ref_genes)
            if ref_count == 0:
                self.ref_genes = self.annotations.genes()
                self.referenceRadioBox.buttons[1].setText("Reference set")
                self.referenceRadioBox.buttons[1].setDisabled(True)
                self.information(
                    2, "Unable to extract gene names from reference dataset. " "Using entire genome for reference"
                )
                self.use_reference_dataset = 0
            else:
                self.referenceRadioBox.buttons[1].setText("Reference set ({} genes)".format(ref_count))
                self.referenceRadioBox.buttons[1].setDisabled(False)
                self.information(2)
        else:
            self.use_reference_dataset = 0
            self.ref_genes = []

        if not self.ref_genes:
            self.error(1, "No valid reference set")
            return {}

        evidences = []
        for etype in go.evidence_types_ordered:
            if self.use_evidence_type[etype]:
                evidences.append(etype)
        aspect = ['Process', 'Component', 'Function'][self.aspect_index]

        self.progressBarInit(processEvents=False)
        self.setBlocking(True)
        self.__state = State.Running

        if self.input_genes:
            f = self._executor.submit(
                self.annotations.get_enriched_terms,
                self.input_genes,
                self.ref_genes,
                evidences,
                aspect=aspect,
                prob=self.probFunctions[self.prob_func],
                use_fdr=False,
                progress_callback=methodinvoke(self, "_progress_bar_set", (float,)),
            )
            fw = FutureWatcher(f, parent=self)
            fw.done.connect(self.__on_enrichment_done)
            fw.done.connect(fw.deleteLater)
            return
        else:
            f = Future()
            f.set_result({})
            self.__on_enrichment_done(f)

    def __on_enrichment_done(self, results):
        # type: (Future[Dict[str, tuple]]) -> None
        self.progressBarFinished(processEvents=False)
        self.setBlocking(False)
        self.setStatusMessage("")
        if self.__state & State.Stale:
            self.__state = State.Ready
            self.__invalidate()
            return

        self.__state = State.Ready
        try:
            results = results.result()  # type: Dict[str, tuple]
        except Exception as ex:
            results = {}
            error = str(ex)
            self.error(1, error)

        if results:
            terms = list(results.items())
            fdr_vals = statistics.FDR([d[1] for _, d in terms])
            terms = [(key, d + (fdr,)) for (key, d), fdr in zip(terms, fdr_vals)]
            terms = dict(terms)

        else:
            terms = {}

        self.terms = terms

        if not self.terms:
            self.warning(0, "No enriched terms found.")
        else:
            self.warning(0)

        self.treeStructDict = {}
        ids = self.terms.keys()

        self.treeStructRootKey = None

        parents = {}
        for _id in ids:
            parents[_id] = {term for _, term in self.ontology[_id].related}

        children = {}
        for term in self.terms:
            children[term] = {id for id in ids if term in parents[id]}

        for term in self.terms:
            self.treeStructDict[term] = TreeNode(self.terms[term], children[term])
            if not self.ontology[term].related and not getattr(self.ontology[term], "is_obsolete", False):
                self.treeStructRootKey = term

        self.set_graph(terms)
        self._update_enrichment_report_output()
        self.commit()

    def _update_enrichment_report_output(self):
        terms = sorted(self.terms.items(), key=lambda item: item[1][1])
        # Create and send the enrichemnt report table.
        terms_domain = Orange.data.Domain(
            [],
            [],
            # All is meta!
            [
                Orange.data.StringVariable("GO Term Id"),
                Orange.data.StringVariable("GO Term Name"),
                Orange.data.ContinuousVariable("Cluster Frequency"),
                Orange.data.ContinuousVariable("Genes in Cluster", number_of_decimals=0),
                Orange.data.ContinuousVariable("Reference Frequency"),
                Orange.data.ContinuousVariable("Genes in Reference", number_of_decimals=0),
                Orange.data.ContinuousVariable("p-value"),
                Orange.data.ContinuousVariable("FDR"),
                Orange.data.ContinuousVariable("Enrichment"),
                Orange.data.StringVariable("Genes"),
            ],
        )

        terms = [
            [
                t_id,
                self.ontology[t_id].name,
                len(genes) / len(self.input_genes),
                len(genes),
                r_count / len(self.ref_genes),
                r_count,
                p_value,
                fdr,
                len(genes) / len(self.input_genes) * len(self.ref_genes) / r_count,
                ",".join(genes),
            ]
            for t_id, (genes, p_value, r_count, fdr) in terms
            if genes and r_count
        ]

        if terms:
            x = numpy.empty((len(terms), 0))
            m = numpy.array(terms, dtype=object)
            terms_table = Orange.data.Table.from_numpy(terms_domain, x, metas=m)
        else:
            terms_table = None
        self.send("Enrichment Report", terms_table)

    @Slot(float)
    def _progress_bar_set(self, value):
        assert QThread.currentThread() is self.thread()
        self.progressBarSet(value, processEvents=None)

    @Slot()
    def _progress_bar_finish(self):
        assert QThread.currentThread() is self.thread()
        self.progressBarFinished(processEvents=None)

    def filter_graph(self, graph):
        if self.filter_by_p_value_nofdr:
            graph = go.filter_by_p_value(graph, self.max_p_value_no_fdr)
        if self.filter_by_p_value:  # FDR
            graph = dict(filter(lambda item: item[1][3] <= self.max_p_value, graph.items()))
        if self.filter_by_num_of_instances:
            graph = dict(filter(lambda item: len(item[1][0]) >= self.min_num_of_instances, graph.items()))
        return graph

    def filter_and_display_graph(self):
        if self.input_data and self.originalGraph is not None:
            self.graph = self.filter_graph(self.originalGraph)
            if self.originalGraph and not self.graph:
                self.warning(1, "All found terms were filtered out.")
            else:
                self.warning(1)
            self.clear_graph()
            self.display_graph()

    def set_graph(self, graph=None):
        self.originalGraph = graph
        if graph:
            self.filter_and_display_graph()
        else:
            self.graph = {}
            self.clear_graph()

    def clear_graph(self):
        self.listView.clear()
        self.listViewItems = []
        self.sigTerms.clear()

    def display_graph(self):
        from_parent_dict = {}
        self.termListViewItemDict = {}
        self.listViewItems = []

        def enrichment(t):
            try:
                return len(t[0]) / t[2] * (len(self.ref_genes) / len(self.input_genes))
            except ZeroDivisionError:
                # TODO: find out why this happens
                return 0

        max_fold_enrichment = max([enrichment(term) for term in self.graph.values()] or [1])

        def add_node(term, parent, parent_display_node):
            if (parent, term) in from_parent_dict:
                return
            if term in self.graph:
                display_node = GOTreeWidgetItem(
                    self.ontology[term],
                    self.graph[term],
                    len(self.input_genes),
                    len(self.ref_genes),
                    max_fold_enrichment,
                    parent_display_node,
                )
                display_node.goId = term
                self.listViewItems.append(display_node)
                if term in self.termListViewItemDict:
                    self.termListViewItemDict[term].append(display_node)
                else:
                    self.termListViewItemDict[term] = [display_node]
                from_parent_dict[(parent, term)] = True
                parent = term
            else:
                display_node = parent_display_node

            for c in self.treeStructDict[term].children:
                add_node(c, parent, display_node)

        if self.treeStructDict:
            add_node(self.treeStructRootKey, None, self.listView)

        terms = self.graph.items()
        terms = sorted(terms, key=lambda item: item[1][1])
        self.sigTableTermsSorted = [t[0] for t in terms]

        self.sigTerms.clear()
        for i, (t_id, (genes, p_value, ref_count, fdr)) in enumerate(terms):
            item = GOTreeWidgetItem(
                self.ontology[t_id],
                (genes, p_value, ref_count, fdr),
                len(self.input_genes),
                len(self.ref_genes),
                max_fold_enrichment,
                self.sigTerms,
            )
            item.goId = t_id

        self.listView.expandAll()
        for i in range(5):
            self.listView.resizeColumnToContents(i)
            self.sigTerms.resizeColumnToContents(i)
        self.sigTerms.resizeColumnToContents(6)
        width = min(self.listView.columnWidth(0), 350)
        self.listView.setColumnWidth(0, width)
        self.sigTerms.setColumnWidth(0, width)

    def view_selection_changed(self):
        if self.selectionChanging:
            return

        self.selectionChanging = 1
        self.selectedTerms = []
        selected = self.listView.selectedItems()
        self.selectedTerms = list({lvi.term.id for lvi in selected})
        self.example_selection()
        self.selectionChanging = 0

    def table_selection_changed(self):
        if self.selectionChanging:
            return

        self.selectionChanging = 1
        self.selectedTerms = []
        selected_ids = {self.sigTerms.itemFromIndex(index).goId for index in self.sigTerms.selectedIndexes()}

        for i in range(self.sigTerms.topLevelItemCount()):
            item = self.sigTerms.topLevelItem(i)
            selected = item.goId in selected_ids
            term = item.goId

            if selected:
                self.selectedTerms.append(term)

            for lvi in self.termListViewItemDict[term]:
                try:
                    lvi.setSelected(selected)
                    if selected:
                        lvi.setExpanded(True)
                except RuntimeError:  # Underlying C/C++ object deleted
                    pass
        self.selectionChanging = 0
        self.example_selection()

    def example_selection(self):
        self.commit()

    def commit(self):
        if self.input_data is None or self.originalGraph is None or self.annotations is None:
            return
        if self.__state & State.Stale:
            return

        terms = set(self.selectedTerms)
        genes = reduce(operator.ior, (set(self.graph[term][0]) for term in terms), set())

        evidences = []
        for etype in go.evidence_types_ordered:
            if self.use_evidence_type[etype]:
                evidences.append(etype)

        all_terms = self.annotations.get_annotated_terms(
            genes, direct_annotation_only=self.selection_direct_annotation, evidence_codes=evidences
        )

        if self.selection_disjoint > 0:
            count = defaultdict(int)
            for term in self.selectedTerms:
                for g in all_terms.get(term, []):
                    count[g] += 1
            ccount = 1 if self.selection_disjoint == 1 else len(self.selectedTerms)
            selected_genes = [gene for gene, c in count.items() if c == ccount and gene in genes]
        else:
            selected_genes = reduce(operator.ior, (set(all_terms.get(term, [])) for term in self.selectedTerms), set())

        if self.use_attr_names:
            selected = [
                column
                for column in self.input_data.domain.attributes
                if self.gene_id_attribute in column.attributes
                and str(column.attributes[self.gene_id_attribute]) in set(selected_genes)
            ]

            domain = Orange.data.Domain(selected, self.input_data.domain.class_vars, self.input_data.domain.metas)
            new_data = self.input_data.from_table(domain, self.input_data)
            self.send("Data on Selected Genes", new_data)

        else:
            selected_rows = []
            for row_index, row in enumerate(self.input_data):
                gene_in_row = str(row[self.gene_id_column])
                if gene_in_row in self.input_genes and gene_in_row in selected_genes:
                    selected_rows.append(row_index)

                if selected_rows:
                    selected = self.input_data[selected_rows]
                else:
                    selected = None

                self.send("Data on Selected Genes", selected)

    def show_info(self):
        dialog = QDialog(self)
        dialog.setModal(False)
        dialog.setLayout(QVBoxLayout())
        label = QLabel(dialog)
        label.setText("Ontology:\n" + self.ontology.header if self.ontology else "Ontology not loaded!")
        dialog.layout().addWidget(label)

        label = QLabel(dialog)
        label.setText(
            "Annotations:\n" + self.annotations.header.replace("!", "")
            if self.annotations
            else "Annotations not loaded!"
        )
        dialog.layout().addWidget(label)
        dialog.show()

    def onDeleteWidget(self):
        """Called before the widget is removed from the canvas."""
        self.annotations = None
        self.ontology = None
        gc.collect()  # Force collection


def fmtp(score):
    return "%0.5f" % score if score > 10e-4 else "%0.1e" % score


def fmtpdet(score):
    return "%0.9f" % score if score > 10e-4 else "%0.5e" % score


class GOTreeWidgetItem(QTreeWidgetItem):
    def __init__(self, term, enrichment_result, n_cluster_genes, n_ref_genes, max_fold_enrichment, parent):
        super().__init__(parent)
        self.term = term
        self.enrichmentResult = enrichment_result
        self.nClusterGenes = n_cluster_genes
        self.nRefGenes = n_ref_genes
        self.maxFoldEnrichment = max_fold_enrichment

        querymapped, pvalue, refmappedcount, fdr = enrichment_result

        querymappedcount = len(querymapped)
        if refmappedcount > 0 and n_ref_genes > 0 and n_cluster_genes > 0:
            enrichment = (querymappedcount / refmappedcount) * (n_ref_genes / n_cluster_genes)
        else:
            enrichment = numpy.nan

        self.enrichment = enrichment

        self.setText(0, term.name)

        fmt = "%" + str(-int(math.log(max(n_cluster_genes, 1)))) + "i (%.2f%%)"
        self.setText(1, fmt % (querymappedcount, 100.0 * querymappedcount / (n_cluster_genes or 1)))

        fmt = "%" + str(-int(math.log(max(n_ref_genes, 1)))) + "i (%.2f%%)"
        self.setText(2, fmt % (refmappedcount, 100.0 * refmappedcount / (n_ref_genes or 1)))

        self.setText(3, fmtp(pvalue))
        self.setToolTip(3, fmtpdet(pvalue))
        self.setText(4, fmtp(fdr))  # FDR
        self.setToolTip(4, fmtpdet(fdr))
        self.setText(5, ", ".join(querymapped))
        self.setText(6, "%.2f" % enrichment)
        self.setToolTip(6, "%.2f" % enrichment)
        self.setToolTip(0, "<p>" + term.__repr__()[6:].strip().replace("\n", "<br>"))
        self.sortByData = [
            term.name,
            querymappedcount,
            refmappedcount,
            pvalue,
            fdr,
            ", ".join(querymapped),
            enrichment,
        ]

    def data(self, col, role):
        if role == Qt.UserRole:
            if self.maxFoldEnrichment > 0:
                return self.enrichment / self.maxFoldEnrichment
            else:
                return numpy.nan
        else:
            return super().data(col, role)

    def __lt__(self, other):
        col = self.treeWidget().sortColumn()
        return self.sortByData[col] < other.sortByData[col]


class EnrichmentColumnItemDelegate(QItemDelegate):
    def paint(self, painter, option, index):
        self.drawBackground(painter, option, index)
        value = index.data(Qt.UserRole)
        if isinstance(value, float) and numpy.isfinite(value):
            painter.save()
            painter.setBrush(QBrush(Qt.blue, Qt.SolidPattern))
            painter.drawRect(
                option.rect.x(), option.rect.y(), int(value * (option.rect.width() - 1)), option.rect.height() - 1
            )
            painter.restore()
        else:
            super().paint(painter, option, index)


if __name__ == "__main__":

    def test_main(argv=sys.argv):
        from AnyQt.QtWidgets import QApplication

        app = QApplication(list(argv))
        argv = app.arguments()
        if len(argv) > 1:
            data = Orange.data.Table(argv[1])
        else:
            data = None

        w = OWGOBrowser()
        w.show()
        w.raise_()
        w.setDataset(data)
        w.handleNewSignals()
        rval = app.exec_()
        w.setDataset(None)
        w.handleNewSignals()
        w.saveSettings()
        w.onDeleteWidget()
        return rval

    sys.exit(test_main())



""" Orange data table and other data manipulation utilities """
import numbers
from types import SimpleNamespace
from typing import Tuple, Sequence
from functools import wraps

from Orange.data import Table, Variable
from Orange.widgets.widget import Msg, OWWidget
from Orange.widgets.utils.messages import UnboundMsg

__missing_annotation = UnboundMsg('Missing annotation on gene IDs and organism in the input data.')
__missing_gene_id = UnboundMsg('Missing gene ID information. Make sure that Table is properly annotated.')
__missing_tax_id = UnboundMsg('Missing organism information. Make sure that Table is properly annotated.')
__unable_to_locate_genes = UnboundMsg('Unable to locate genes. Make sure that Table is properly annotated.')


class TableAnnotation(SimpleNamespace):
    """ Data Table hints """

    # Organism in data table
    tax_id: str = 'taxonomy_id'

    # This indicates position of genes in data table
    gene_as_attr_name: str = 'gene_as_attribute_name'

    # This indicates a column name (if genes are in rows)
    gene_id_column: str = 'gene_id_column'

    # This indicates attribute name (if genes are in columns)
    gene_id_attribute: str = 'gene_id_attribute'


def check_table_annotation(f):
    """Wrapper for widget's input method that checks if the data on the input is correctly annotated.

    A widget in bioinformatics add-on expects that every Table has additional
    information stored as table attributes:
       - taxonomy_id = 'taxonomy id for given organism'
       - gene_as_attribute_name = 'location of gene names (rows/columns)'
       - gene_id_attribute/gene_id_column = 'attribute/column name'

    """

    @wraps(f)
    def wrapper(widget, data: Table, *args, **kwargs):
        widget.Error.add_message('missing_annotation', __missing_annotation)
        widget.Error.add_message('missing_gene_id', __missing_gene_id)
        widget.Error.add_message('missing_tax_id', __missing_tax_id)
        widget.Error.add_message('unable_to_locate_genes', __unable_to_locate_genes)

        widget.Error.missing_annotation.clear()
        widget.Error.missing_gene_id.clear()
        widget.Error.missing_tax_id.clear()
        widget.Error.unable_to_locate_genes.clear()

        if data is not None and isinstance(data, Table):
            attributes: dict = data.attributes

            tax_id: str = TableAnnotation.tax_id
            gene_id_column: str = TableAnnotation.gene_id_column
            gene_id_attribute: str = TableAnnotation.gene_id_attribute
            gene_as_attr_name: str = TableAnnotation.gene_as_attr_name

            if not attributes:
                widget.Error.missing_annotation()
                data = None

            elif tax_id not in attributes:
                widget.Error.missing_tax_id()
                data = None

            elif gene_as_attr_name not in attributes:
                widget.Error.unable_to_locate_genes()
                data = None

            elif gene_as_attr_name in attributes:
                if (attributes[gene_as_attr_name] and gene_id_attribute not in attributes) or (
                    not attributes[gene_as_attr_name] and gene_id_column not in attributes
                ):

                    widget.Error.unable_to_locate_genes()
                    data = None

        return f(widget, data, *args, **kwargs)

    return wrapper


# TODO: remove this and replace with TableAnnotation namespace

# species
TAX_ID = 'taxonomy_id'

# Will be set to True if gene names are represented as attribute names.
# If gene names are in rows, we set this value to False. (user must select proper column index)
GENE_AS_ATTRIBUTE_NAME = 'gene_as_attribute_name'

# Name of the column where rows are gene ids
GENE_ID_COLUMN = 'gene_id_column'

# Name of the variable attribute that holds gene id
GENE_ID_ATTRIBUTE = 'gene_id_attribute'

# Error strings
ERROR_ON_MISSING_ANNOTATION = 'Missing annotation on gene IDs and organism in the input data.'
ERROR_ON_MISSING_GENE_ID = 'Missing gene ID information in the input data'
ERROR_ON_MISSING_TAX_ID = 'Missing organism information in the input data'



"""
This module cluster the projection of data (usually it is 2D projection) with
one of the standard algorithms and attach a certain number of labels per
cluster.

Example:

>>> from Orange.projection import TSNE
>>> from Orange.data import Table
>>> from orangecontrib.bioinformatics.utils import serverfiles
>>> from orangecontrib.bioinformatics.annotation.annotate_projection import \
...     annotate_projection
>>> from orangecontrib.bioinformatics.annotation.annotate_samples import \
...     AnnotateSamples
>>>
>>> # load data
>>> data = Table("https://datasets.orange.biolab.si/sc/aml-1k.tab.gz")
>>> marker_p = serverfiles.localpath_download(
...     'marker_genes','panglao_gene_markers.tab')
>>> markers = Table(marker_p)
>>>
>>> # annotate data with labels
>>> annotator = AnnotateSamples()
>>> annotations = annotator.annotate_samples(data, markers)
>>>
>>> # project data in 2D
>>> tsne = TSNE(n_components=2)
>>> tsne_model = tsne(data)
>>> embedding = tsne_model(data)
>>>
>>> # get clusters and annotations for clusters
>>> clusters, clusters_meta, eps = annotate_projection(annotations, embedding,
...     clustering_algorithm=DBSCAN, eps=1.2)

In case when user uses a DBSCAN algorithm and do not provide eps to the
`annotate_projection` function it is computed automatically with knn method.

>>> clusters, clusters_meta, eps = annotate_projection(annotations, embedding,
...     clustering_algorithm=DBSCAN)

"""
from math import sqrt
from bisect import bisect_left, bisect_right
from collections import Counter

import numpy as np
import pyclipper
from scipy.spatial import Delaunay, distance

from Orange.data import Table, Domain, DiscreteVariable
from Orange.clustering import DBSCAN


def cluster_data(coordinates, clustering_algorithm=DBSCAN, **kwargs):
    """
    This function receives data and cluster them.

    Parameters
    ----------
    coordinates : Orange.data.Table
        Visualisation coordinates - embeddings
    clustering_algorithm : callable
        Algorithm used for clustering.

    Returns
    -------
    Orange.data.Table
        List of cluster indices.
    """
    learner = clustering_algorithm(**kwargs)
    clusters = learner(coordinates)
    if not isinstance(clusters, np.ndarray):  # old clustering method
        clusters = clusters(coordinates)
        clusters = np.array(
            list(
                map(int, map(clusters.domain.attributes[0].repr_val, clusters.X[:, 0]))
            )
        ).flatten()

    # sort classes in descending order base on number of cases in the cluster
    sorted_clust_idx = [v for v, _ in Counter(clusters).most_common() if v != -1]

    # re-indexed array
    new_clustering = np.empty(len(clusters))
    new_clustering[:] = np.nan  # nan for not clustered
    # reindex based on descending cluster size
    for i, v in enumerate(sorted_clust_idx):
        new_clustering[clusters == v] = i

    # create the table
    new_domain = Domain(
        [
            DiscreteVariable(
                "Clusters",
                values=["C{}".format(i) for i in range(1, len(sorted_clust_idx) + 1)],
            )
        ]
    )
    return Table(new_domain, new_clustering.reshape((-1, 1)))


def assign_labels(clusters, annotations, labels_per_cluster):
    """
    This function assigns a certain number of labels per cluster. Each cluster
    gets `labels_per_cluster` number of most common labels in cluster assigned.

    Parameters
    ----------
    clusters : Orange.data.Table
        Cluster indices for each item.
    annotations : Orange.data.Table
        Table with annotations and their probabilities.
    labels_per_cluster : int
        Number of labels that need to be assigned to each cluster.

    Returns
    -------
    dict
        Dictionary with cluster index as a key and list of annotations as a
        value. Each list include tuples with the annotation name and their
        proportion in the cluster.
    Orange.data.Table
        The array with the annotation assigned to the item.
    """
    clusters_unique = set(clusters.domain[0].values)

    if len(annotations.domain) == 0:
        return {}, Table(
            Domain([DiscreteVariable("Annotation", values=[])]),
            np.ones((len(clusters), 1)) * np.nan,
        )

    labels = np.array(list(map(str, annotations.domain.attributes)))

    # remove rows with all nans
    nan_mask = np.isnan(annotations.X).all(axis=1)
    ann_not_nan = annotations.X[~nan_mask]

    # find indices and labels
    annotation_best_idx = np.nanargmax(ann_not_nan, axis=1)
    annotation_best = labels[annotation_best_idx]

    # join back together
    items_annotations = np.empty(annotations.X.shape[0], dtype=labels.dtype)
    items_annotations[~nan_mask] = annotation_best

    annotations_clusters = {}
    for cl in clusters_unique:
        mask = (
            np.array(
                list(map(clusters.domain.attributes[0].repr_val, clusters.X[:, 0]))
            ).flatten()
            == cl
        )
        labels_cl = items_annotations[mask]
        # remove nans from labels
        labels_cl_filtered = labels_cl[~(labels_cl == "")]

        counts = Counter(labels_cl_filtered)
        common_labels = counts.most_common(labels_per_cluster)

        if len(common_labels) > 0:
            annotations_clusters[cl] = [
                (label, c / len(labels_cl)) for label, c in common_labels
            ]

    # pack item annotations to Table
    nan_mask = items_annotations == ""
    values, indices = np.unique(items_annotations[~nan_mask], return_inverse=True)
    corrected_idx = np.ones(items_annotations.shape) * np.nan
    corrected_idx[~nan_mask] = indices
    domain = Domain([DiscreteVariable("Annotation", values=values)])
    item_annotations = Table.from_list(domain, corrected_idx.reshape((-1, 1)))

    return annotations_clusters, item_annotations


def labels_locations(coordinates, clusters):
    """
    Function computes the location of the label for each cluster.
    The location is compute as a center point.

    Parameters
    ----------
    coordinates : Orange.data.Table
        Visualisation coordinates - embeddings
    clusters : Orange.data.Table
        Cluster indices for each item.

    Returns
    -------
    dict
        The coordinates for locating the label. Dictionary with cluster index
        as a key and tuple (x, y) as a value.
    """
    clusters_unique = set(clusters.domain[0].values) - {"-1"}  # -1 is not clustered
    locations = {}
    for cl in clusters_unique:
        mask = (
            np.array(
                list(map(clusters.domain.attributes[0].repr_val, clusters.X[:, 0]))
            ).flatten()
            == cl
        )
        cl_coordinates = coordinates.X[mask, :]
        x, y = 1 / 2 * (np.min(cl_coordinates, axis=0) + np.max(cl_coordinates, axis=0))
        locations[cl] = (x, y)
    return locations


def get_epsilon(coordinates, k=10, skip=0.1):
    """
    The function computes the epsilon parameter for DBSCAN through method
    proposed in the paper.

    Parameters
    ----------
    coordinates : Orange.data.Table
        Visualisation coordinates - embeddings
    k : int
        Number kth observed neighbour
    skip : float
        Percentage of skipped neighborus.

    Returns
    -------
    float
        Epsilon parameter for DBSCAN
    """
    x = coordinates.X
    if len(x) > 1000:  # subsampling is required
        i = len(x) // 1000
        x = x[::i]

    d = distance.squareform(distance.pdist(x))
    k = min(k + 1, len(coordinates) - 1)
    kth_point = np.argpartition(d, k, axis=1)[:, k]
    # k+1 since first one is item itself
    kth_dist = np.sort(d[np.arange(0, len(kth_point)), kth_point])

    # currently mark proportion equal to skip as a noise
    return kth_dist[-int(np.round(len(kth_dist) * skip))]


def _angle(v1, v2):
    """
    Compute clockwise angles between v1 and v2. Both vectors are
    given as a tuple with two points ([x1, y1], [x2, y2]) such that
    [x2, y2] of v1 == [x1, y1] of v2.
    The angle is given in degrees between 0 and 2 * pi
    """
    v1, v2 = np.array(v1), np.array(v2)
    x1, y1 = v1[0] - v1[1]
    x2, y2 = v2[1] - v2[0]
    dot = np.sum(x1 * x2 + y1 * y2)  # dot product
    det = x1 * y2 - y1 * x1  # determinant
    angle = np.arctan2(det, dot)  # atan2(y, x) or atan2(sin, cos)
    return -angle if angle <= 0 else np.pi + np.pi - angle


def _find_hull(edges_list, points_list, starting_edge):
    """
    This function return a single hull which starts and ends in
    the starting_edge.

    Parameters
    edges_list : list
        List of edges. Each edge is presented as a tuple of two indices which
        tell the starting and ending note. The index correspond to location
        of point in points_list. This list must be sorted in the ascending
        order.
    points_list : list
        List of points location. Each point has x and y location.
    starting_edge : int
        The index of the list where hull starts.

    Returns
    -------
    np.ndarray
        The array with the hull/polygon points
    list
        List of booleans that indicates whether each point was used or not in
        the resulting polygon.
    """
    firsts = [x[0] for x in edges_list]  # first elements for bisection
    used_edges = [False] * len(edges_list)
    used_edges[starting_edge] = True

    # remember start and next point
    # start point is required for stop condition
    start, next_point = edges_list[starting_edge]

    # remember current polygon around points
    poly = [points_list[start], points_list[next_point]]

    # we count number of steps to stop iteration in case it is locked
    # in some dead cycle. It can be a result of some unexpected cases.
    count = 0
    while start != next_point and count < len(edges_list):
        # find the index where the first value equal to next_point
        # appear
        ind_left = bisect_left(firsts, next_point)
        # find the index next to the last value
        ind_right = bisect_right(firsts, next_point)

        # check if there are more edges available from the same point
        if ind_right - ind_left > 1:
            # select the most distant one in clockwise direction
            # it is probably the point on the outer hull
            # with this we prevent a hull to discover cycles inside
            # a polygon
            ang = -1
            for i in range(ind_left, ind_right):
                cur_ang = _angle(
                    (poly[-2], poly[-1]), (poly[-1], points_list[edges_list[i][1]])
                )
                if cur_ang > ang:
                    ang = cur_ang
                    ind_left = i
        # save a next point of the polgon
        used_edges[ind_left] = True
        next_point = edges_list[ind_left][1]
        poly.append(points_list[next_point])
        count += 1
    return np.array(poly), used_edges


def edges_to_polygon(edges_list, points_list):
    """
    This function connect edges in polygons. It computes all possible hulls -
    yes some clusters have more of them when they have a hole in the middle.
    It then selects one that is outer hull.

    Parameters
    ----------
    edges_list : list
        List of edges. Each edge is presented as a tuple of two indices which
        tell the starting and ending note. The index correspond to location
        of point in points_list
    points_list : list
        List of points location. Each point has x and y location.

    Returns
    -------
    np.ndarray
        The array with the hull/polygon points.
    """
    # sort based on first element of tuple to enable bisection search
    edges_list = sorted(edges_list, key=lambda x: x[0])
    # need to use all edges
    used = [False] * len(edges_list)

    # it is possible that we will find more separate hulls -
    # it happen in cases when a polygon has inner cycles
    polygons = []
    while not all(used):
        i = used.index(False)
        poly, new_used = _find_hull(edges_list, points_list, i)
        polygons.append(poly)
        used = [u1 or u2 for u1, u2 in zip(used, new_used)]

    # select polygon that is outside - the widest and the highest is the
    # most outside
    height_width = [np.sum(p.max(axis=0) - p.min(axis=0)) for p in polygons]
    i = height_width.index(max(height_width))

    return polygons[i]


def compute_concave_hulls(coordinates, clusters, epsilon):
    """
    Function computes the points of the concave hull around points.

    Parameters
    ----------
    coordinates : Orange.data.Table
        Visualisation coordinates - embeddings
    clusters : Orange.data.Table
       Cluster indices for each item.
    epsilon : float
        Epsilon used by DBSCAN to cluster the data

    Returns
    -------
    dict
       The points of the concave hull. Dictionary with cluster index
       as a key and np.ndaray of points as a value -
       [[x1, y1], [x2, y2], [x3, y3], ...]
    """

    def get_shape(pts, eps):
        """
        Compute the shape (concave hull) of a set of a cluster.
        """

        def add_edge(edges_list, i, j):
            """
            Add a line between the i-th and j-th points,
            if not in the list already. Remove the lines that are not outer
            edges - when (j, i) already in list means that two triangles
            has same edge what means it is not an outer edge
            """
            if (j, i) in edges_list:
                # if both neighboring triangles are in shape, it's not a
                # boundary edge
                edges_list.remove((j, i))
                return
            edges_list.add((i, j))

        if len(pts) < 4:
            # When you have a triangle, there is no sense in computing the hull
            rng = list(range(3))
            return edges_to_polygon(zip(rng, rng[1:] + rng[:1]), pts)

        tri = Delaunay(pts)
        edges = set()
        # loop over triangles:
        # ia, ib, ic = indices of corner points of the triangle
        for ia, ib, ic in tri.vertices:
            pa = pts[ia]
            pb = pts[ib]
            pc = pts[ic]

            # Lengths of sides of triangle
            a = sqrt((pa[0] - pb[0]) ** 2 + (pa[1] - pb[1]) ** 2)
            b = sqrt((pb[0] - pc[0]) ** 2 + (pb[1] - pc[1]) ** 2)
            c = sqrt((pc[0] - pa[0]) ** 2 + (pc[1] - pa[1]) ** 2)

            # filter - longest edge of triangle smaller than epsilon
            if max(a, b, c) <= eps:
                add_edge(edges, ia, ib)
                add_edge(edges, ib, ic)
                add_edge(edges, ic, ia)

        polygon = edges_to_polygon(edges, pts)
        return polygon

    hulls = {}
    clusters_array = np.array(
        list(map(clusters.domain.attributes[0].repr_val, clusters.X[:, 0]))
    )
    for cl in set(clusters_array) - {"None", "?"}:
        points = coordinates.X[clusters_array == cl]

        # subsample when more than 1000 points
        # it keeps time finding hull under 0.3 s on my computer
        if points.shape[0] > 1000:
            points = points[np.random.randint(points.shape[0], size=1000), :]

        # compute the concave hul
        hull = get_shape(points, eps=epsilon * 2)

        # epsilon seems to be good parameter for lines to be smooth enough
        # selecting epsilon for the distance
        # shows approximately what is DBSCAN neighbourhood
        # those line buffer the line for 3 * epsilon and move it back for
        # 2 * epsilon. The it will make a hull more smooth.

        # pyclipper work with integer so points need to be scaled first to keep
        # precision. It is undo with scal_from_clipper
        scaling_factor = 1000
        scaled_hull = pyclipper.scale_to_clipper(hull, scaling_factor)

        # buffer the hull fro epsilon * 3
        pco1 = pyclipper.PyclipperOffset()
        pco1.AddPath(scaled_hull, pyclipper.JT_ROUND, pyclipper.ET_CLOSEDPOLYGON)
        im_solution = pco1.Execute(epsilon * scaling_factor * 3)
        # buffer the hull fro epsilon * -2
        pco2 = pyclipper.PyclipperOffset()
        pco2.AddPath(im_solution[0], pyclipper.JT_ROUND, pyclipper.ET_CLOSEDPOLYGON)
        solution = pyclipper.scale_from_clipper(
            pco2.Execute(epsilon * scaling_factor * (-2)), scaling_factor
        )

        hulls[cl] = np.array(solution).reshape(-1, 2)
    return hulls


def _filter_clusters(clusters, clusters_meta):
    """
    Function removes clusters that does not has any labels
    """
    clust_map = {
        c: i
        for i, c in enumerate(
            c for c in clusters.domain["Clusters"].values if c in clusters_meta
        )
    }

    # change cluster indices
    clust_idx = np.zeros(clusters.X.shape) * np.nan
    for i, cl in enumerate(
        map(clusters.domain.attributes[0].repr_val, clusters.X[:, 0])
    ):
        clust_idx[i, 0] = clust_map.get(cl, np.nan)
    new_clusters = Table(
        Domain(
            [
                DiscreteVariable(
                    "Clusters",
                    values=clusters.domain["Clusters"].values[: len(clust_map)],
                )
            ]
        ),
        clust_idx,
    )
    # change cluster names in metas
    new_clusters_meta = {
        "C{}".format(clust_map[cl] + 1): v for cl, v in clusters_meta.items()
    }
    return new_clusters, new_clusters_meta


def annotate_projection(
    annotations,
    coordinates,
    clustering_algorithm=DBSCAN,
    labels_per_cluster=3,
    **kwargs
):
    """
    Function cluster the data based on coordinates, and assigns a certain number
    of labels per cluster. Each cluster gets `labels_per_cluster` number of most
    common labels in cluster assigned.

    Parameters
    ----------
    annotations : Orange.data.Table
        Table with annotations and their probabilities.
    coordinates : Orange.data.Table
        Visualisation coordinates - embeddings
    clustering_algorithm : callable, optional (default = DBSCAN)
        Algorithm used in clustering.
    labels_per_cluster : int, optional (default = 3)
        Number of labels that need to be assigned to each cluster.

    Returns
    -------
    Orange.data.Table
        List of cluster indices.
    dict
        Dictionary with cluster index as a key and list of annotations as a
        value. Each list include tuples with the annotation name and their
        proportion in the cluster.
    dict
        The coordinates for locating the label. Dictionary with cluster index
        as a key and tuple (x, y) as a value.
    """
    assert len(annotations) == len(
        coordinates
    ), "Number of coordinates does not match to number of annotations"
    # sklearn clustering want to have one example
    assert len(coordinates) > 0, "At least one data point need to be provided"
    assert (
        len(coordinates.domain) > 0
    ), "Coordinates need to have at least one attribute"

    eps = kwargs.get("eps", get_epsilon(coordinates))
    if clustering_algorithm == DBSCAN:
        kwargs["eps"] = eps

    # get clusters
    clusters = cluster_data(coordinates, clustering_algorithm, **kwargs)

    # assign top n labels to group
    annotations_cl, item_annotations = assign_labels(
        clusters, annotations, labels_per_cluster
    )

    labels_loc = labels_locations(coordinates, clusters)

    concave_hull = compute_concave_hulls(coordinates, clusters, eps)

    # crate the dictionary with annotations, labels locations, and hulls for
    # each cluster
    clusters_meta = {}
    # note clusters without labels are not added to the dictionary
    for cl in annotations_cl.keys():
        clusters_meta[cl] = (annotations_cl[cl], labels_loc[cl], concave_hull[cl])

    # we do not want to color clusters that does not have labels so we
    # remove clusters without labels from the table
    clusters, clusters_meta = _filter_clusters(clusters, clusters_meta)

    # add the labels to the cluster table
    clusters_ann = Table(
        Domain(clusters.domain.attributes + item_annotations.domain.attributes),
        np.concatenate((clusters.X, item_annotations.X), axis=1),
    )

    return clusters_ann, clusters_meta, eps


def cluster_additional_points(coordinates, hulls, cluster_attribute=None):
    """
    This function receives additional points and assign them current existing
    clusters based on current concave hull.

    Parameters
    ----------
    coordinates : Orange.data.Table
        Visualisation coordinates - embeddings
    hulls : dict
        Concave hull for each cluster
    cluster_attribute : Orange.data.DiscreteVariable (optional)
        A variable for clusters. If cluster_attribute is provided it will be
        used in the creation of the resulting Table.

    Returns
    -------
    Orange.data.Table
        Cluster label for each point
    """

    def point_in_polygon_test(test_point, polygon_points):
        """
        This function uses the horizontal ray casting to find out if the point
        is in the hull/polygon. For each point, it tests how many times the
        horizontal ray from test_point to infinity crosses the polygon edge. If
        it happens odd many times the point is in the polygon.
        https://stackoverflow.com/a/2922778/3551700
        """
        test_x = test_point[0]
        test_y = test_point[1]
        # flipping bool from True to False is similar to counting odd numbers
        # of intersections. If it will be True at the end odd number of
        # intersections happened
        is_inside = False

        for (x1, y1), (x2, y2) in zip(
            polygon_points,
            np.concatenate((polygon_points[1:], polygon_points[:1]), axis=0),
        ):
            # ray crosses the edge if test_y between both y from an edge
            # and if intersection on the right of the test_x
            if (y1 > test_y) != (y2 > test_y):
                # compute the intersection between the horizontal ray and
                # polygon edge
                intersection_x = (x2 - x1) * (test_y - y1) / (y2 - y1) + x1
                if test_x < intersection_x:
                    is_inside = not is_inside
        return is_inside

    clusters = [None] * len(coordinates)
    for cluster, hull in hulls.items():
        for i, c in enumerate(coordinates.X):
            if point_in_polygon_test(c, hull):
                clusters[i] = cluster

    if cluster_attribute is not None:
        assert all(
            i in cluster_attribute.values for i in set(clusters) - {None}
        ), "cluster_attribute does not have all required values."
    # create the table
    new_domain = Domain(
        [
            DiscreteVariable("Clusters", values=sorted(hulls.keys()))
            if cluster_attribute is None
            else cluster_attribute
        ]
    )
    return Table(
        new_domain, np.array(list(map(new_domain[0].to_val, clusters))).reshape(-1, 1)
    )


if __name__ == "__main__":
    # run hull creation at Iris data
    data = Table("iris")[:, 2:4]
    clustered_data = Table(
        Domain([DiscreteVariable("cl", values=["1", "2", "3"])]),
        [[0]] * 50 + [[1]] * 50 + [[2]] * 50,
    )
    compute_concave_hulls(data, clustered_data, epsilon=0.5)
